__author__ = 'admin'

# import lxml.etree as etree
from FastOMA.zoo.familyanalyzer.orthoxmlquery import OrthoXMLQuery
import FastOMA.zoo.familyanalyzer as fa

# To parse the orthoxml and get the related etree object: etree.parse(fn_orthoxml) denote here by XML
# To get the root of your etree object: etree_object.getroot() denote here by XML_root
# The orthoxml is using two ids:
#   - The internal id denotes by the tag "id" which is found in the mapping in the beginning of the orthoxml and inside the orthologous groups
#   - The external id denotes by the tag "protId"(sometimes with the tag "geneId" or BOTH) which is found only in the begining of the orthoxml

def get_mapping_orthoxml(XML_root):

    map_ext_2_int = {}
    map_int_2_ext = {}

    for el in OrthoXMLQuery.getInputGenes(XML_root, species=None):
        int_id = None
        ext_id = []
        for _idtag,_id in el.attrib.items():
            if _idtag == "id":
                int_id = _id
            else:
                ext_id.append(str(_id))

        if int_id != None and ext_id != []:
            map_ext_2_int[frozenset(ext_id)] = str(int_id)
            map_int_2_ext[str(int_id)] = ext_id

    return map_ext_2_int, map_int_2_ext

def get_all_hogs(XML_root):
    '''
    Return a list of HOGs (etree object)
    '''
    return OrthoXMLQuery.getToplevelOrthologGroups(XML_root)

def get_hog_by_id(XML_root, hog_id):
    '''
    Return the HOG of interest (etree object)
    '''
    for OG_etree in OrthoXMLQuery.getToplevelOrthologGroups(XML_root):
        if int(OG_etree.get("id")) == int(hog_id):
            return OG_etree

def get_hogs_by_id(XML_root, list_hog_id):
    '''
    Return HOGs of interest (etree object)
    '''
    list_hog = []
    for OG_etree in OrthoXMLQuery.getToplevelOrthologGroups(XML_root):
        if int(OG_etree.get("id")) in list_hog_id:
            list_hog.append(OG_etree)
    return list_hog


def get_hogs_with_query_genes(XML_root, list_genes_you_want_to_look, use_internal_id = False, return_hog_etree = False):
    '''
    XML_root: root of the etree element
    list_genes_you_want_to_look: list of the genes external id
    use_internal_id: if the list of genes given is composed of internal id
    return_hog_etree: return the etree element of the hod instead of the id
    ext_id: external id use to parse the orthoxml

    return: each gene with its related HOG id (dictionnary)

    Protips: list(set(gene_with_its_HOG.values())) will return the list of hogs return

    '''
    map_ext_2_int, map_int_2_ext = get_mapping_orthoxml(XML_root)
    gene_with_its_HOG = {} # keys:gene & values:hog

    if use_internal_id == False:
        list_genes_int = []
        for ext_id in list_genes_you_want_to_look:
            for ext_id_keys in map_ext_2_int.keys():
                if str(ext_id) in ext_id_keys:
                    list_genes_int.append(str(map_ext_2_int[ext_id_keys]))
    else:
        list_genes_int = []
        for intid in list_genes_you_want_to_look:
            list_genes_int.append(str(intid))
    for OG_etree in get_all_hogs(XML_root):
        OG_genes = [ gene_etree for gene_etree in OG_etree.getiterator() if gene_etree.tag == "{http://orthoXML.org/2011/}geneRef" ]
        for gene_etree in OG_genes:
            if str(gene_etree.get("id")) in list_genes_int:
                if return_hog_etree:
                    gene_with_its_HOG[str(map_int_2_ext[gene_etree.get("id")])]=OG_etree
                else:
                    gene_with_its_HOG[str(map_int_2_ext[gene_etree.get("id")])]=OG_etree.get("id")
    return gene_with_its_HOG

def get_size_all_HOGs(XML_root):
    '''
    XML_root: root of the etree element
    return: dictionary key:hog_id and value:number of genes
    '''
    OGs = {}
    for OG_etree in get_all_hogs(XML_root):
        OG_genes = [ gene_etree for gene_etree in OG_etree.getiterator() if gene_etree.tag == "{http://orthoXML.org/2011/}geneRef" ]
        OGs[OG_etree.get("id")]= len(OG_genes)
    return OGs


def get_speciesTree(fn_orthoxml):
    '''
    Extract from a orthoxml the species tree based on the internal topology
    :param XML: orthoxml file
    :return: newick string
    '''
    op = fa.OrthoXMLParser(fn_orthoxml)
    tax = fa.TaxonomyFactory.newTaxonomy(op)
    op.augmentTaxonomyInfo(tax)
    return tax.newick()


def get_list_species(XML_root):
    '''
    return a list of species contains into an orthoxml.

    :XML_root: root of the etree element
    :return: list of species
    '''
    species = [ e.get("name") for e in XML_root.getiterator() if e.tag == "{http://orthoXML.org/2011/}species" ]
    return species


import unittest

import FastOMA._config
from Bio import AlignIO
from FastOMA._wrappers import infer_gene_tree
from FastOMA.zoo.wrappers import WrapperError
import pathlib
this_dir = pathlib.Path(__file__).parent


class FastTreeTester(unittest.TestCase):
    def test_failing_tree_building_reports_error_from_fasttree(self):
        msa = AlignIO.read(this_dir / "data" / "failing-msa.fa", "fasta")
        with self.assertLogs("FastOMA", level="ERROR") as cm:
            with self.assertRaises(WrapperError):
                infer_gene_tree(msa, "/tmp")
            self.assertIn("Non-unique name", "\n".join(cm.output))

    def test_treebuilding_with_correct_msa(self):
        msa = AlignIO.read(this_dir / "data" / "correct-msa.fa", "fasta")
        tree = infer_gene_tree(msa, "/tmp")
        self.assertIn("HUMAN01350", tree)


import itertools
import time
import os
from Bio import AlignIO, SeqIO
from six import StringIO
from ..abstract_cli import AbstractCLI
from .base_aligner import Aligner, AlignmentInput, DataType
from ...seq_utils.utils import iter_seqrecs_from_any
from ...wrappers import WrapperError
from ..options import StringOption, FlagOption, IntegerOption, FloatOption, MultiOption, OptionSet
import tempfile
import logging



logger = logging.getLogger(__name__)


class MafftCLI(AbstractCLI):
    """
    Mafft low-level command line interface

    :Example:

    ::

        mafft_cli = _MafftCLI()
        process = mafft_cli(cmd='mafft args...')
        stdout = mafft_cli.get_stdout()
    """

    @property
    def _default_exe(self):
        return 'mafft'

    # def _set_help(self):
    #     self(help=True, wait=True)
    #     self._help = self.get_stdout()


def set_default_dna_options(aligner):
    """
    Dummy function as sensible default already provided by mafft --auto
    """
    aligner.options = get_default_options()
    aligner.options['--auto'].set_value(True)
    aligner.options['--nuc'].set_value(True)


def set_default_protein_options(aligner):
    """
    Dummy function as sensible default already provided by mafft --auto
    """
    aligner.options = get_default_options()
    aligner.options['--auto'].set_value(True)
    aligner.options['--amino'].set_value(True)


class Mafft(Aligner):
    """
    Convenient wrapper for Mafft multiple sequence aligner

    The wrapper is written as a callable class.
    This can hold data (state) to do with the operation it performs, so it can keep results,
    execution times and other metadata, as well as perform the task.

    This is a basic implementation that can be extended. The important parts are
    __init__ (does the setup) and __call__ (does the work). All
    else are helper methods.

    :Example:

    ::

        callable_wrapper = Mafft(aln)
        result = callable_wrapper()
        time_taken = callable_wrapper.elapsed_time
        result_again = callable_wrapper.result


    .. note:: There exists an ipython notebook on how to work with wrappers,
         including dealing with non-default parameters.
    """

    def __init__(self, input_, *args, **kwargs):
        super(Mafft, self).__init__(input_, *args, **kwargs)
        self.options = get_default_options()
        if self.datatype == DataType.DNA:
            set_default_dna_options(self)
        else:
            set_default_protein_options(self)

    def __call__(self, *args, **kwargs):
        """
        Anything to do with calling Mafft should go here.
        If any extra arguments need to be passed they can
        be specified (listed as *args and **kwargs for now).
        """
        start = time.time()  # time the execution

        if self.input_type == AlignmentInput.OBJECT:  # different operation depending on what it is
            with tempfile.TemporaryDirectory() as tmpdir:
                if self.options['--merge'].active and not os.path.exists(self.options['--merge'].get_value()):
                    self.input, copy_input = itertools.tee(self.input)
                    with open(os.path.join(tmpdir, "submsa_table.txt"), 'wt') as fsub:
                        seq_idx= 1
                        for msa in copy_input:
                            assert isinstance(msa, AlignIO.MultipleSeqAlignment)
                            line = " ".join([str(c + seq_idx) for c in range(len(msa))])
                            fsub.write(line + "\n")
                            seq_idx = seq_idx + len(msa)

                    self.options['--merge'].set_value(fsub.name)
                with open(os.path.join(tmpdir, "sequences.fa"), 'wt') as fh:
                    SeqIO.write(iter_seqrecs_from_any(self.input), fh, 'fasta')
                output, error = self._call(fh.name, *args, **kwargs)
        else:
            output, error = self._call(self.input, *args, **kwargs)

        logger.debug('Output of Mafft: stdout={}; stderr={}'.format(output, error))
        if len(output) == 0 and len(error) > 0:
            logger.warning('is MAFFT_BINARIES set correctly: {}'.format(os.getenv('MAFFT_BINARIES', '')))
            raise WrapperError('Mafft did not compute any alignments. StdErr: {}'.format(error))
        self.result = self._read_result(output)  # store result
        self.stdout = output
        self.stderr = error

        end = time.time()
        self.elapsed_time = end - start
        return self.result
        # End call

    # Any other accessory methods 
    def _call(self, filename, *args, **kwargs):
        """
        Call underlying low level _Mafft wrapper. 
        Options are passed via *args and **kwargs
        [This only covers the simplest automatic
         case]
        """
        self.cli('{} {}'.format(self.command(), filename),
                 wait=True)
        return self.cli.get_stdout(), self.cli.get_stderr()

    def command(self):
        return str(self.options)

    def _read_result(self, output):
        """
        Read back the result.
        """
        fileobj = StringIO(output)
        return AlignIO.read(fileobj, 'fasta')

    def _init_cli(self, binary):
        return MafftCLI(executable=binary)


def get_default_options():
    return OptionSet([
        # Algorithm

        # Automatically selects an appropriate strategy from L-INS-i, FFT-NS-i
        # and FFT-NS-2, according to data size. Default: off (always FFT-NS-2)
        FlagOption('--auto', False, active=False),

        # Distance is calculated based on the number of shared 6mers. Default: on
        FlagOption('--6merpair', True, active=False),

        # All pairwise alignments are computed with the Needleman-Wunsch algorithm.
        # More accurate but slower than --6merpair. Suitable for a set of globally
        # alignable sequences. Applicable to up to ~200 sequences. A combination
        # with --maxiterate 1000 is recommended (G-INS-i). Default: off
        # (6mer distance is used)
        FlagOption('--globalpair', False, active=False),

        # All pairwise alignments are computed with the Smith-Waterman algorithm.
        # More accurate but slower than --6merpair. Suitable for a set of locally
        # alignable sequences. Applicable to up to ~200 sequences. A combination
        # with --maxiterate 1000 is recommended (L-INS-i). Default: off
        # (6mer distance is used)
        FlagOption('--localpair', False, active=False),

        # All pairwise alignments are computed with a local algorithm with the
        # generalized affine gap cost (Altschul 1998). More accurate but slower than
        # --6merpair. Suitable when large internal gaps are expected. Applicable to
        # up to ~200 sequences. A combination with --maxiterate 1000 is recommended
        # (E-INS-i). Default: off (6mer distance is used)
        FlagOption('--genafpair', False, active=False),

        # All pairwise alignments are computed with FASTA (Pearson and Lipman 1988).
        # FASTA is required. Default: off (6mer distance is used)
        FlagOption('--fastapair', False, active=False),

        # Weighting factor for the consistency term calculated from pairwise
        # alignments. Valid when either of --blobalpair, --localpair, --genafpair,
        # --fastapair or --blastpair is selected. Default: 2.7
        FloatOption('--weighti', 2.7, active=False),

        # Guide tree is built number times in the progressive stage. Valid with 6mer
        # distance. Default: 2
        IntegerOption('--retree', 2, active=False),

        # number cycles of iterative refinement are performed. Default: 0
        IntegerOption('--maxiterate', 0, active=False),

        # Use FFT approximation in group-to-group alignment. Default: on
        FlagOption('--fft', True, active=False),

        # Do not use FFT approximation in group-to-group alignment. Default: off
        FlagOption('--nofft', False, active=False),

        # Alignment score is not checked in the iterative refinement stage. Default:
        # off (score is checked)
        FlagOption('--noscore', False, active=False),

        # Use the Myers-Miller (1988) algorithm. Default: automatically turned on
        # when the alignment length exceeds 10,000 (aa/nt).
        FlagOption('--memsave', False, active=False),

        # Use a fast tree-building method (PartTree, Katoh and Toh 2007) with the
        # 6mer distance. Recommended for a large number (> ~10,000) of sequences are
        # input. Default: off
        FlagOption('--parttree', False, active=False),

        # The PartTree algorithm is used with distances based on DP. Slightly more
        # accurate and slower than --parttree. Recommended for a large number
        # (> ~10,000) of sequences are input. Default: off
        FlagOption('--dpparttree', False, active=False),

        # The PartTree algorithm is used with distances based on FASTA. Slightly
        # more accurate and slower than --parttree. Recommended for a large number
        # (> ~10,000) of sequences are input. FASTA is required. Default: off
        FlagOption('--fastaparttree', False, active=False),

        # The number of partitions in the PartTree algorithm. Default: 50
        IntegerOption('--partsize', 50, active=False),

        # Do not make alignment larger than number sequences. Valid only with the
        # --*parttree options. Default: the number of input sequences
        IntegerOption('--groupsize', 1, active=False),

        # Parameter

        # Gap opening penalty at group-to-group alignment. Default: 1.53
        FloatOption('--op', 1.53, active=False),

        # Offset value, which works like gap extension penalty, for group-to-group
        # alignment. Deafult: 0.123
        FloatOption('--ep', 0.123, active=False),

        # Gap opening penalty at local pairwise alignment. Valid when the
        # --localpair or --genafpair option is selected. Default: -2.00
        FloatOption('--lop', -2.0, active=False),

        # Offset value at local pairwise alignment. Valid when the --localpair or
        # --genafpair option is selected. Default: 0.1
        FloatOption('--lep', 0.1, active=False),

        # Gap extension penalty at local pairwise alignment. Valid when the
        # --localpair or --genafpair option is selected. Default: -0.1
        FloatOption('--lexp', -0.1, active=False),

        # Gap opening penalty to skip the alignment. Valid when the --genafpair
        # option is selected. Default: -6.00
        FloatOption('--LOP', -6.00, active=False),

        # Gap extension penalty to skip the alignment. Valid when the --genafpair
        # option is selected. Default: 0.00
        FloatOption('--LEXP', 0.0, active=False),

        # BLOSUM number matrix (Henikoff and Henikoff 1992) is used. number=30, 45,
        # 62 or 80. Default: 62
        IntegerOption('--bl', 62, active=False),

        # JTT PAM number (Jones et al. 1992) matrix is used. number>0.
        # Default: BLOSUM62
        IntegerOption('--jtt', 100, active=False),

        IntegerOption('--randomseed', 1234, active=True),


        # Transmembrane PAM number (Jones et al. 1994) matrix is used. number>0.
        # Default: BLOSUM62
        IntegerOption('--tm', 100, active=False),

        # Use a user-defined AA scoring matrix. The format of matrixfile is the same
        # to that of BLAST. Ignored when nucleotide sequences are input.
        # Default: BLOSUM62
        StringOption('--aamatrix', '', active=False),

        # Incorporate the AA/nuc composition information into the scoring matrix.
        # Deafult: off
        FlagOption('--fmodel', False, active=False),

        # Output

        # Output format: clustal format. Default: off (fasta format)
        FlagOption('--clustalout', False, active=False),

        # Output order: same as input. Default: on
        FlagOption('--inputorder', True, active=False),

        # Output order: aligned. Default: off (inputorder)
        FlagOption('--reorder', False, active=False),

        # Guide tree is output to the input.tree file. Default: off
        FlagOption('--treeout', False, active=False),

        # Do not report progress. Default: off
        FlagOption('--quiet', False, active=False),

        # Input

        # Assume the sequences are nucleotide. Default: auto
        FlagOption('--nuc', False, active=False),

        # Assume the sequences are amino acid. Default: auto
        FlagOption('--amino', False, active=False),

        # use unusual characters (e.g., U as selenocysteine in protein sequence; i as inosine in nucleotide sequence), Default=False
        FlagOption('--anysymbol', False, active=False),

        # Seed alignments given in alignment_n (fasta format) are aligned with
        # sequences in input. The alignment within every seed is preserved.
        MultiOption('--seed', None, active=False),

        # Choose analysis mode automatically (default=True)
        FlagOption('--auto', True, active=True),

        # Merge two or more sub MSAs into a single MSA
        # Two or more sub MSAs are merged into a single MSA. Sub MSAs are assumed
        # to be phylogenetically separated from each other.
        StringOption('--merge', '', active=False),
        IntegerOption('--thread', -1, active=False),
    ])





import os

from . import _utils_subhog
from . import _infer_subhog
from . import _config
from ._config import logger_hog


def fastoma_infer_subhogs():
    _config.set_configs()
    logger_hog.debug("logger_level is "+str(_config.logger_level))
    logger_hog.debug("input_rhog_folder is " + str(_config.input_rhog_folder))
    logger_hog.debug("parallel is " + str(_config.parallel))
    logger_hog.debug("species_tree_checked is " + str(_config.species_tree_checked))
    logger_hog.debug("fragment_detection is " + str(_config.fragment_detection))
    logger_hog.debug("low_so_detection is " + str(_config.low_so_detection))
    logger_hog.debug("inferhog_max_workers_num is " + str(_config.inferhog_max_workers_num))
    logger_hog.debug("inferhog_tresh_ratio_gap_row is " + str(_config.inferhog_tresh_ratio_gap_row))
    logger_hog.debug("inferhog_tresh_ratio_gap_col is " + str(_config.inferhog_tresh_ratio_gap_col))
    logger_hog.debug("inferhog_min_cols_msa_to_filter is " + str(_config.inferhog_min_cols_msa_to_filter))
    logger_hog.debug("big_rhog_size is " + str(_config.big_rhog_size))
    logger_hog.debug("omamer_family_threshold is " + str(_config.omamer_family_threshold))
    logger_hog.debug("hogclass_max_num_seq is " + str(_config.hogclass_max_num_seq))
    logger_hog.debug("hogclass_min_cols_msa_to_filter is " + str(_config.hogclass_min_cols_msa_to_filter))
    logger_hog.debug("inferhog_resume_rhog is " + str(_config.inferhog_resume_rhog))
    logger_hog.debug("inferhog_resume_subhog is " + str(_config.inferhog_resume_subhog))
    logger_hog.debug("fragment_detection is " + str(_config.fragment_detection))
    logger_hog.debug("fragment_detection_msa_merge is " + str(_config.fragment_detection_msa_merge))
    logger_hog.debug("low_so_detection is " + str(_config.low_so_detection))
    logger_hog.debug("threshold_dubious_sd is " + str(_config.threshold_dubious_sd))
    logger_hog.debug("overlap_fragments is " + str(_config.overlap_fragments))
    logger_hog.debug("gene_trees_write is " + str(_config.gene_trees_write))
    logger_hog.debug("msa_write is " + str(_config.msa_write))
    logger_hog.debug("gene_trees_write_all is " + str(_config.gene_trees_write_all))
    logger_hog.debug("msa_write_all is " + str(_config.msa_write_all))
    logger_hog.debug("keep_subhog_each_pickle is " + str(_config.keep_subhog_each_pickle))


    address_rhogs_folder = _config.input_rhog_folder
    # inferhog_concurrent_on = False
    inferhog_concurrent_on = _config.parallel #  sys.argv[2]   # "False"  # "False"  #

    if inferhog_concurrent_on:
        print("parallelization for subhog inference is on.")

    pickles_rhog_folder = "./pickle_hogs" # pickles_temp/ pickle_rhogs
    if not os.path.exists(pickles_rhog_folder):
        os.makedirs(pickles_rhog_folder)

    pickles_subhog_folder_all = "./"
    print("input is", address_rhogs_folder)

    list_rhog_fastas_files = _utils_subhog.list_rhog_fastas(address_rhogs_folder)
    print("there are ", len(list_rhog_fastas_files), "rhogs in the input folder")

    rhogs_fa_folder = address_rhogs_folder

    list_rhog_fastas_files_rem = _utils_subhog.list_rhog_fastas(address_rhogs_folder)
    print("there are ", len(list_rhog_fastas_files_rem), "rhogs remained in the input folder", list_rhog_fastas_files_rem[:5] )

    hogs_rhog_xml_batch = _infer_subhog.read_infer_xml_rhogs_batch(list_rhog_fastas_files_rem, inferhog_concurrent_on, pickles_rhog_folder, pickles_subhog_folder_all, rhogs_fa_folder)

    print("finsihed ", address_rhogs_folder)

# TODO all below

"""
  
- use same seed for hog sampling and fasttree/mafft if they have 
- add python code for validate the input make sure the file is there and decent
- check input tree 
_utils_subhog line 92 add check for spaces or chars in 

- print as debug all the variables 
# [_config.oma_database_address, _config.working_folder_root , _config.species_tree_address , _config.working_id ,
 _config.protein_format_qfo_dataset, _config.in_folder, _config.omamer_fscore_treshold_big_rhog,
  _config.treshold_big_rhog_szie, _config.gene_trees_write, _config.keep_subhog_each_pickle, 
  _config.hogclass_max_num_seq, _config.hogclass_min_cols_msa_to_filter, _config.hogclass_tresh_ratio_gap_col, _config.automated_trimAL, _config.lable_SD_internal ,
   _config.rooting_method, _config.rooting_mad_executable_path , _config.inferhog_tresh_ratio_gap_row , _config.inferhog_tresh_ratio_gap_col , _config.inferhog_min_cols_msa_to_filter
    , _config.inferhog_filter_all_msas_row , _config.inferhog_resume_rhog  , _config.inferhog_resume_subhog , _config.inferhog_max_workers_num , _config.inferhog_min_hog_size_xml, _config.logger_level]

FileExistsError: [Errno 17] File exists: '/work/FAC/FBM/DBC/cdessim2/default/smajidi1/fastget/qfo3/working_nf//pickles_subhog/'

when concurrent has a issue it doesnt stop
Out[6]: {<Future at 0x7f1b48d9afa0 state=finished raised TypeError>: 'KORCO_'}
add eception to show , whn this happens for which taxnomic level and rhog

proteins were associated with known HOGs, and warn if less than ~80 % of proteins are associated with known HOGs.

double check 
     prots_to_remove
     it seems that it always an input argument! 
     

"""




import itertools
from abc import ABCMeta, abstractmethod
from future.utils import with_metaclass
import dendropy as dpy
import lxml.etree

__all__ = ["iter_pairwise_relations"]


class BaseRelationExtractor(with_metaclass(ABCMeta, object)):
    def __init__(self, obj):
        self.obj = obj
        self._rel_type = 'ortholog'
        self.relations = []

    def default_node_list(self):
        return self.obj

    def extract_pairwise_relations(self, node=None, rel_type='ortholog'):
        self.check_rel_type(rel_type)

        root_nodes = [node]
        if node is None:
            root_nodes = self.default_node_list()

        for group in root_nodes:
            self.relations = []
            self._extract_pw(group)
            for rel in self.relations:
                yield rel

    @abstractmethod
    def is_leaf(self, node):
        pass

    def is_internal_node(self, node):
        return not self.is_leaf(node)

    @abstractmethod
    def leaf_label(self, leaf):
        pass

    def _extract_pw(self, node):
        if self.is_leaf(node):
            return {self.leaf_label(node)}
        elif self.is_internal_node(node):
            nodes_of_children = [self._extract_pw(child) for child in self.get_children(node)]
            if self.shall_extract_from_node(node):
                for child1, child2 in itertools.combinations(nodes_of_children, 2):
                    for gId1, gId2 in itertools.product(child1, child2):
                        self.relations.append((gId1, gId2))
            nodes = set.union(*nodes_of_children)
            return nodes
        else:
            return set([])

    @abstractmethod
    def shall_extract_from_node(self, node):
        pass

    @abstractmethod
    def get_children(self, node):
        """return an iterable over the children nodes of `node`"""
        pass

    def check_rel_type(self, rel_type):
        if callable(rel_type):
            self.shall_extract_from_node = rel_type
        elif rel_type not in ('ortholog', 'paralog'):
            raise ValueError('rel_type argument needs to be one of "ortholog" or "paralog"')
        else:
            self._rel_type = rel_type


class OrthoXmlRelationExtractor(BaseRelationExtractor):
    def __init__(self, fh, id_attribute=None, **kwargs):
        super(OrthoXmlRelationExtractor, self).__init__(fh)
        if id_attribute is not None:
            self.geneRef_to_id = {}
            ns = "{http://orthoXML.org/2011/}"
            for g in self.obj.findall(f'./{ns}species/{ns}database/{ns}genes/{ns}gene'):
                try:
                    self.geneRef_to_id[g.get('id')] = g.get(id_attribute)
                except AttributeError:
                    raise KeyError(f"gene {g} does not have an attribute '{id_attribute}'")
        else:
            self.geneRef_to_id = None

    def default_node_list(self):
        return self.obj.find(".//{http://orthoXML.org/2011/}groups")

    def shall_extract_from_node(self, node):
        return lxml.etree.QName(node).localname == self._rel_type + "Group"

    def is_internal_node(self, node):
        return lxml.etree.QName(node).localname in ('orthologGroup', 'paralogGroup')

    def leaf_label(self, leaf):
        res = leaf.get('id')
        if self.geneRef_to_id:
            res = self.geneRef_to_id[res]
        return res

    def is_leaf(self, node):
        return lxml.etree.QName(node).localname == "geneRef"

    def get_children(self, node):
        return node


class TreeRelationExtractor(BaseRelationExtractor):
    def shall_extract_from_node(self, node):
        annos = set(map(node.annotations.get_value, ('Ev', 'D')))
        annos.remove(None)
        if self._rel_type == 'ortholog':
            if len(annos) == 0 or 'S' in annos or 'F' in annos:
                return True
        elif self._rel_type == 'paralog':
            if 'D' in annos or 'T' in annos:
                return True
        return False

    def get_children(self, node):
        return node.child_node_iter()

    def is_leaf(self, node):
        return node.is_leaf()

    def leaf_label(self, leaf):
        return leaf.taxon.label

    def default_node_list(self):
        return [self.obj.seed_node]

    def __init__(self, obj):
        super(TreeRelationExtractor, self).__init__(obj)


def iter_pairwise_relations(obj, rel_type=None, node=None, **kwargs):
    """iterate over the induced pairwise relations from `obj`.

    This function extracts all the induced pairwise relations of type
    `rel_type` from a the hierarchical structure passed as parameter
    `obj`. As hierarchical structures the method accepts either an
    orthoxml file handle or a labeled :class:`dendropy.datamodel.treemodel.Tree`

    rel_type can be used to specify which type of relations should be
    returned. by default, it returns the orthologous relations. The
    argument can be set to 'ortholog', 'paralog' or - if obj is a
    dendropy tree - it can also be a function. This function takes
    as input an internal tree node and should return True iff the
    relations induced by this node should be taken.

    :param obj: a reference to an orthoxml file opened with lxml or a
        dendropy tree.

    :param rel_type: either 'ortholog', 'paralog'. if obj is a
        dendropy tree, rel_type can also or a function

    :param node: a reference to the root node of the substructure
        from which the pairwise relations should be extracted.
        Defaults to the root of the tree or the <groups> tag
        respectively.

    :param id_attribute: applies only for orthoxml files. If parameter
        is provided, instead of the internal geneRef-id, the value that
        is stored in the gene under the given attribute is returned. So
        meaningful values can be `protId` or `geneId`.

    :param kwargs: additional arguments that are passed. So far
        none will be used. Foreseen examples would be a callback
        function on leaf nodes.

    :return: an iterator yielding tuples of leaf ids of the requested
        evolutionary type.

    :Example:

        >>> tree = dpy.Tree.get_from_string(
        ...       '(1,(2,(3,4))[&&NHX:Ev=D],(5,6));', schema='newick')
        >>> sorted(list(iter_pairwise_relations(tree, rel_type='paralog')))
        [('2', '3'), ('2', '4')]
    """
    if hasattr(obj, 'docinfo') and obj.docinfo.root_name == "orthoXML":
        parser = OrthoXmlRelationExtractor(obj, **kwargs)
    elif isinstance(obj, dpy.Tree):
        parser = TreeRelationExtractor(obj)

    else:
        raise ValueError('cannot extract pairwise relations from obj')

    for rel in parser.extract_pairwise_relations(node=node,
                                                 rel_type=rel_type if rel_type is not None else 'ortholog'):
        yield rel



import shutil
from . import _config
from pathlib import Path
logger = _config.logger_hog


class BatchBuilder:
    def __init__(self, outdir: Path, max_size: int):
        self.outdir = outdir
        self.max_size = max_size

    def __enter__(self):
        self.cur_batch = []
        self.cur_size = 0
        self.counter = 0
        self.outdir.mkdir(parents=True, exist_ok=True)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if len(self.cur_batch) > 0:
            self._flush()

    def add_hog(self, hog_file: Path):
        self.cur_batch.append(hog_file)
        self.cur_size += hog_file.stat().st_size
        logger.debug("adding %s with size %d to batch %d", hog_file, hog_file.stat().st_size, self.counter)
        if self.cur_size > self.max_size:
            self._flush()
            self.counter += 1

    def _flush(self):
        batch_dir = self.outdir / str(self.counter)
        batch_dir.mkdir()
        for fn in self.cur_batch:
            shutil.copy(fn, batch_dir)
        logger.debug("creating batch %s with %d families; total size of files is %d",
                     batch_dir, len(self.cur_batch), self.cur_size)
        self.cur_size = 0
        self.cur_batch = []


def folder_1h_rhog(roothog_path: Path, output_folder_big: Path, output_folder_rest: Path):
    # create a list of hogs in descending filesize order
    hog_size_tuples = sorted([(f, f.stat().st_size) for f in roothog_path.rglob("*.fa")], key=lambda x: -x[1])
    with BatchBuilder(output_folder_big, 1) as big_hogs, \
            BatchBuilder(output_folder_rest, _config.sum_list_rhogs_filesize_thresh) as rest_hogs:
        for hog, fsize in hog_size_tuples:
            if fsize > _config.big_rhog_filesize_thresh:
                big_hogs.add_hog(hog)
            else:
                rest_hogs.add_hog(hog)


def fastoma_batch_roothogs():
    import argparse
    parser = argparse.ArgumentParser(description="Analyse roothog families and create batches for analysis")
    parser.add_argument('--input-roothogs', required=True, help="folder where input roothogs are stored")
    parser.add_argument('--out-big', required=True, help="folder where the big single family hogs should be stored")
    parser.add_argument('--out-rest', required=True, help="folder where the remaining families should be stored in"
                                                          "batch subfolder structure.")
    parser.add_argument('-v', default=0, action="count", help="incrase verbosity")
    conf = parser.parse_args()
    logger.setLevel(level=30 - 10 * min(conf.v, 2))

    folder_1h_rhog(Path(conf.input_roothogs), Path(conf.out_big), Path(conf.out_rest))




import os
import time
import logging
import random
from pyparsing import ParseException
import shutil
from Bio import AlignIO, SeqIO

from .base_treebuilder import TreeBuilder, AlignmentInput, DataType
from .parsers import RaxmlParser

from ..abstract_cli import AbstractCLI
from ..options import StringOption, FlagOption, IntegerOption, FloatOption, MultiOption, OptionSet

from ...file_utils import TempFile,TempDir

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.INFO)


class RaxmlCLI(AbstractCLI):
    @property
    def _default_exe(self):
        return ['raxmlHPC','raxmlHPC-PTHREADS']


def set_default_dna_options(treebuilder):
    """
    Dummy function as sensible default
    """
    treebuilder.options = get_default_options()


def set_default_protein_options(treebuilder):
    """
    Dummy function as sensible default
    """
    treebuilder.options = get_default_options()


class Raxml(TreeBuilder):

    def __init__(self, alignment, *args, **kwargs):
        self.options = get_default_options()
        super(Raxml, self).__init__(alignment=alignment, *args, **kwargs)
        if self.input is not None:
            if self.datatype == DataType.DNA:
                set_default_dna_options(self)
            else:
                set_default_protein_options(self)



    def __call__(self, *args, **kwargs):
        """
        Sets up temporary output files and calls raxml using _call() function.
        Writes temporary input file if we're working with SeqIO object
        Saves the stdout and stderr and returns
        """
        start = time.time()  # time the execution

        #Need to create temp directory to put raxml output here
        with TempDir() as tmpd:
            if self.input_type is AlignmentInput.OBJECT:  # different operation depending on what it is
                with TempFile() as filename:
                    SeqIO.write(self.input, filename, 'phylip-relaxed') # default interleaved
                    output, error = self._call(filename,tmpd, *args, **kwargs)
            elif self.input_type is AlignmentInput.FILENAME:
                filename = self.input
                output, error = self._call(filename, tmpd, *args, **kwargs)
            else:
                output, error = self._call(None,tmpd, *args, **kwargs)
            self.result = self._read_result(tmpd)  # store result
        self.stdout = output
        self.stderr = error

        end = time.time()
        self.elapsed_time = end - start
        return self.result
        # End call

    # Any other accessory methods
    def _call(self, filename, tmpd, *args, **kwargs):
        """
        Call underlying low level _Raxml wrapper.
        Options are passed via *args and **kwargs
        [This only covers the simplest automatic
         case]
        """
        #hard code tmp_output as the output name since we don't save it anyway
        self.cli('{} -n tmp_output -w {tmp_path} -s {seqfile}'.format(self.command(), tmp_path=tmpd, seqfile=filename),
                wait=True)
        return self.cli.get_stdout(), self.cli.get_stderr()

    def command(self):
        return str(self.options)

    def _read_result(self, tmpd):
        """
        Read back the result.
        """

        expected_outfiles = [os.path.join(tmpd, 'RAxML_info.tmp_output'), os.path.join(tmpd, 'RAxML_bestTree.tmp_output')]


        parser = RaxmlParser()

        try:
            if self.options['-f'].get_value() is not '':
                f_value = os.path.splitext(os.path.basename(self.options['-f'].get_value()))[0]

                result = parser.to_dict(*expected_outfiles, dash_f=f_value)
            else:
                result = parser.to_dict(*expected_outfiles, dash_f=None)

        except IOError as ioerr:
            logger.error('Error reading results')
            result = None
        except ParseException as parseerr:
            logger.error('Other parse error', parseerr)
            result = None

        return result

    def _init_cli(self, binary):
        return RaxmlCLI(executable=binary)


def get_default_options():
    return OptionSet([
        # Algorithm

        # Set the model for either DNA or AA alignment
        StringOption('-m', 'PROTGAMMAGTR', active=True),

        # Number of replicates
        IntegerOption('-p', 12345, active=True),

        # If set to true will assume sequential format
        FlagOption('-q', False, active=False),

        # Turn on bootstrapping - set seed
        IntegerOption('-b', 0, active=False),

        # Number of replicates
        IntegerOption('-#', 0, active=False),

        # Turn on rapid bootstrap - specify seed
        IntegerOption('-x', 0, active=False),

        # Sed number of bootstrap replicates
        IntegerOption('-N', 0, active=False),

        # Set number of threads
        IntegerOption('-T', 0, active=False),

        # Tree topology search operation option
        StringOption('-s', 'NNI', active=False),

        # Select algorithm
        StringOption('-f', '', active=False),

        # Specify starting tree
        StringOption('-t', '', active=False),

        # Specify filename of file containing multiple trees
        StringOption('-z', '', active=False),

    ])



import collections

"""UnionFind.py

Union-find data structure. Based on Josiah Carlson's code,
http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/215912
with significant additional changes by D. Eppstein and
Adrian Altenhoff.
"""


class UnionFind(object):
    """Union-find data structure.

    Each unionFind instance X maintains a family of disjoint sets of
    hashable objects, supporting the following two methods:

    - X[item] returns a name for the set containing the given item.
      Each set is named by an arbitrarily-chosen one of its members; as
      long as the set remains unchanged it will keep the same name. If
      the item is not yet part of a set in X, a new singleton set is
      created for it.

    - X.union(item1, item2, ...) merges the sets containing each item
      into a single larger set.  If any item is not yet part of a set
      in X, it is added to X as one of the members of the merged set.
    """

    def __init__(self, elements=None):
        """Create a new union-find structure.

        If elements is not None, the structure gets initialized
        with each element as a singleton component.

        :param elements: an iterable to initialize the structure.
        """

        self.weights = {}
        self.parents = {}
        if elements is not None:
            for elem in iter(elements):
                self.parents[elem] = elem
                self.weights[elem] = 1

    def __getitem__(self, obj):
        """return the name of set which contains obj.

        :param obj: the query object

        :SeeAlso: :meth:`find`"""
        return self.find(obj)

    def find(self, obj):
        """Find and return the name of the set containing the obj.

        If the object is not found in any set, a new singleton set
        is created that holds only this object until it is further merged."""

        # check for previously unknown obj. If unknown, add it
        # as a new cluster
        if obj not in self.parents:
            self.parents[obj] = obj
            self.weights[obj] = 1
            return obj

        # find path of objects leading to the root
        path = [obj]
        root = self.parents[obj]
        while root != path[-1]:
            path.append(root)
            root = self.parents[root]

        # compress the path and return
        for ancestor in path:
            self.parents[ancestor] = root
        return root

    def remove(self, obj):
        """Remove an object from the sets.

        Removes an object entirly from the datastructure. The
        containing set will shrink by this one element.

        :Note: If one tries to accessed it afterwards using
            :meth:`find`, it will be created newly and put as a
            singleton.
        """
        if obj not in self.parents:
            return
        comp = self.find(obj)
        self.weights[comp] -= 1
        self.parents.pop(obj)

    def __iter__(self):
        """Iterate through all items ever found or unioned by this structure."""
        return iter(self.parents)

    def union(self, *objects):
        """Find the sets containing the objects and merge them.

        any number of objects can be passed to this method and
        all of them will be merged into one set containing at
        least these objects.

        :param objects: the objects to be merged. they have to be all
            hashable. If they haven't been initialy added to the UnionFind
            datastructre at instantiation time, they are added at this point
            in time.
        """
        roots = [self[x] for x in objects]
        heaviest = max([(self.weights[r], r) for r in roots], key=lambda x: x[0])[1]
        for r in roots:
            if r != heaviest:
                self.weights[heaviest] += self.weights[r]
                self.parents[r] = heaviest

    def get_components(self):
        """return a list of sets corresponding to the connected
        components of the structure."""
        comp_dict = collections.defaultdict(set)
        for elem in iter(self):
            comp_dict[self[elem]].add(elem)
        comp = list(comp_dict.values())
        return comp



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import numbers

from future.builtins import range
from future.builtins import next
from future.builtins import dict
from future.builtins import zip
from future.builtins import open
from future.builtins import map
from future.builtins import str
from future import standard_library
standard_library.install_hooks()
from past.builtins import basestring

import copy
import io
import itertools
import os
import re
from collections import deque
from functools import reduce

from .orthoxmlquery import OrthoXMLQuery
from .newick import NewickLexer, Streamer
from .tools import PROGRESSBAR, Queue, setup_progressbar


class TaxonomyInconsistencyError(Exception):
    pass


class ParseError(Exception):
    pass


class LinearTaxonomyException(Exception):
    pass


class Taxonomy(object):
    def __init__(self):
        raise NotImplementedError("abstract class")

    def __iter__(self):
        return self.hierarchy[self.root].iter_preorder()

    def __getitem__(self, node_name):
        return self.hierarchy[node_name]

    def finialize_init(self):
        self.extract_descendent_species()
        self.extract_younger_nodes()

    def iterParents(self, node, stopBefore=None):
        """iterates over all the taxonomy nodes towards the root
        which are above 'node' and below 'stopBefore'."""

        if node == stopBefore:
            return
        tn = self.hierarchy[node]
        while tn.up is not None and tn.up.name != stopBefore:
            tn = tn.up
            yield tn.name

    def is_ancestor_of(self, anc, desc):
        """ Returns True if `anc' is an ancestor of `desc'"""
        return anc in self.iterParents(desc)

    def levels_between(self, anc, desc):
        """ Counts the number of levels between anc and desc
        If anc == desc, this is 0. If anc is the immediate parent
        of desc this is 1, grandparent 2 etc. Return -1 if anc is
        not an ancestor of desc."""
        if anc == desc:
            return 0
        if not self.is_ancestor_of(anc, desc): # inefficient, but not *too* inefficient (?)
            return -1
        else:
            parents = list(self.iterParents(desc, stopBefore=anc))
            return len(parents) + 1

    def retain(self, leaves):
        """ Returns a deepcopy of self with all leaves not in `leaves'
        pruned away """
        keep = set(leaves)
        for leaf in leaves:
            keep.update(set(self.iterParents(leaf)))
        lose=set(self.hierarchy.keys()).difference(keep)
        return self._prune_set_of_nodes(lose)

    def prune(self, leaves):
        """ Returns a deepcopy of self with `leaves' pruned away """
        all_leaves = {n.name for n in self.hierarchy[self.root].iter_leaves()}
        keep = all_leaves.difference(leaves)
        return self.retain(keep)

    def _prune_set_of_nodes(self, leaves):
        """ Returns a deepcopy of self with `leaves' pruned away """
        newtax = copy.deepcopy(self)
        for loser in leaves:
            if loser in newtax.hierarchy:
                node = newtax.hierarchy[loser]
                node.up.down.remove(node)
                node.up = None
                del newtax.hierarchy[loser]
        return newtax

    def _countParentAmongLevelSet(self, levels):
        """helper method to count for each level how many levels
        are parent levels. e.g. (arrow: is-parent-of)
          A->B->C
              \>D->E
        will return A=0,B=1,C=2,D=2,E=3
        Cached to speed up multiple calls"""
        if not hasattr(self, '_cpals_cache'):
            self._cpals_cache = {}
        levelSet = frozenset(levels)
        try:
            return self._cpals_cache[levelSet]
        except KeyError:
            counts = dict()
            for lev in levelSet:
                t = set(self.iterParents(lev)).intersection(levelSet)
                counts[lev] = len(t)
            self._cpals_cache[levelSet] = counts
            return counts

    def map_potential_internal_speciesname_to_leaf(self, species_name, code):
        if len(self.descendents[species_name]) == 1:
            return species_name
        elif code is not None:
            return code
        code_candidates = []
        for leaf in self.descendents[species_name]:
            if len(leaf) == 5 and re.match(r'[A-Z][A-Z[0-9]{4}', leaf) is not None:
                code_candidates.append(leaf)
        if len(code_candidates) == 1:
            return code_candidates[0]
        raise TaxonomyInconsistencyError("cannot find a 5letter code species for internal species name")

    def mrca(self, species):
        """Returns most recent common ancestor (MRCA) of a set of species
           This is cached to speed up multiple calls """
        if not hasattr(self, '_mrca_cache'):
            self._mrca_cache = {}
        species = frozenset(species)
        try:
            return self._mrca_cache[species]
        except KeyError:
            if len(species) == 1:
                mrca, = species
                self._mrca_cache[species] = mrca
                return mrca
            ancestors = [set(self.iterParents(s)) for s in species]
            common_ancestors = reduce(lambda x, y: x & y, ancestors)
            mrca = self.mostSpecific(common_ancestors)
            self._mrca_cache[species] = mrca
            return mrca

    def mostSpecific(self, levels):
        """returns the most specific (youngest) level among a set of
        levels. it is required that all levels are on one monophyletic
        lineage, otherwise an Exception is raised."""
        # count how often each element is a child of any other one.
        # the one with len(levels)-1 is the most specific level
        counts = self._countParentAmongLevelSet(levels)
        for lev, count in counts.items():
            if count == len(levels)-1:
                return lev
        raise Exception("Non of the element is subelement of all others")

    def mostGeneralLevel(self, levels):
        """returns the most general (oldest) level among a set of levels."""
        # count who often each element is a child of any other one.
        # the one with len(levels)-1 is the most specific level
        counts = self._countParentAmongLevelSet(levels)
        for lev, count in counts.items():
            if count == 0:
                return lev
        raise Exception("Non of the element is the root of all others")

    def younger_than_filter(self, levels, oldest_permitted):
        """
        Filters a set of levels, removing any that are older than the
        oldest_permitted (oldest_permitted=string: node_name)
        """
        # try:
        #     oldest_permitted_node = self.hierarchy[oldest_permitted]
        # except KeyError:
        #     raise Exception('No node with name {} found in '
        #                     'Taxonomy'.format(oldest_permitted))
        # permitted = [node.name
        #              for node in oldest_permitted_node.iter_preorder()]

        return [lev for lev in levels if lev in self.younger_nodes[oldest_permitted]]

    def printSubTreeR(self, fd, lev=None, indent=0):
        if lev is None:
            lev = self.root
        fd.write("{}{}\n".format(" "*2*indent, lev))
        for child in self.hierarchy[lev].down:
            self.printSubTreeR(fd, child.name, indent+1)

    def get_histories(self, parser, verbosity=0):

        histories = {}

        if PROGRESSBAR and verbosity > 0:
            pbar = setup_progressbar('Getting histories', len(self.hierarchy))
            pbar.start()

        for i, level in enumerate(self.hierarchy, start=1):
            history = parser.getFamHistory()
            history.analyzeLevel(level)
            histories[level] = history
            self.hierarchy[level].attach_fam_history(history)
            if PROGRESSBAR and verbosity > 0:
                pbar.update(i)

        if PROGRESSBAR and verbosity > 0:
            pbar.finish()

        self.histories = histories
        return histories

    def get_comparisons(self, parser, verbosity=0):

        if getattr(self, 'histories', None) is None:
            self.get_histories(parser)

        comparisons = {}
        to_compare = [(node, child) for node in self
                                    for child in node.down]

        if PROGRESSBAR and verbosity > 0:
            pbar = setup_progressbar('Comparing', len(to_compare))
            pbar.start()

        for i, (parent, child) in enumerate(to_compare, start=1):
            parent_history = self.histories[parent.name]
            child_history = self.histories[child.name]
            parent_child_comparison = parent_history.compare(child_history)
            comparisons[(parent.name, child.name)] = parent_child_comparison
            child.attach_level_comparison_result(parent_child_comparison)
            if PROGRESSBAR and verbosity > 0:
                pbar.update(i)

        if PROGRESSBAR and verbosity > 0:
            pbar.finish()

        return comparisons

    def newick(self):
        return str(self.hierarchy[self.root]) + ';'

    def __str__(self):
        fd = io.StringIO()
        self.printSubTreeR(fd)
        res = fd.getvalue()
        fd.close()
        return res

    def extract_descendent_species(self):
        """
        Caches some frequently looked-up information - descendent leaves of every node
        """
        self.descendents = {}
        for k, v in self.hierarchy.items():
            self.descendents[k] = set(l.name for l in v.iter_leaves())

    def extract_younger_nodes(self):
        """
        Caches some frequently looked-up information - descendent nodes of every node
        """
        self.younger_nodes = {}
        for k, v in self.hierarchy.items():
            self.younger_nodes[k] = set(n.name for n in v.iter_preorder())


class LinearTaxonomy(Taxonomy):
    """ Linear taxonomy """

    def __init__(self, taxonomy, histories, comparisons=None):
        # member variable error-checking and setup
        self.check_histories(taxonomy, histories)
        if comparisons is None:
            comparisons = self.generate_comparisons(histories)
        self.check_comparisons(comparisons)
        self.check_comparisons_and_histories(histories, comparisons)

        # build taxonomy
        self.hierarchy = dict()
        self.histories = dict()
        nodes = [TaxNode(h.analyzedLevel) for h in histories]
        self.root = nodes[0].name
        for i in range(len(histories) - 1):
            nodes[i].add_child(nodes[i+1])
            nodes[i+1].add_parent(nodes[i])
            self.hierarchy[nodes[i].name] = nodes[i]
            self.histories[nodes[i].name] = histories[i]
            nodes[i].attach_fam_history(histories[i])
            nodes[i+1].attach_level_comparison_result(comparisons[i])
        nodes[-1].attach_fam_history(histories[-1])
        self.hierarchy[nodes[-1].name] = nodes[-1]
        self.histories[nodes[-1].name] = histories[-1]

    def generate_comparisons(self, histories):
        comparisons = list()
        for i in range(len(histories) - 1):
            comparisons.append(histories[i].compare(histories[i+1]))
        return comparisons

    def check_histories(self, taxonomy, histories):
        labels = [h.analyzedLevel for h in histories]
        for i in range(len(labels) - 1):
            if not taxonomy.is_ancestor_of(labels[i], labels[i+1]):
                raise LinearTaxonomyException('Histories are not in a valid '
                                              'taxonomic sequence')

    def check_comparisons(self, comparisons):
        for i in range(len(comparisons) - 1):
            if not comparisons[i].lev2 == comparisons[i+1].lev1:
                raise LinearTaxonomyException('Comparisons do not overlap')

    def check_comparisons_and_histories(self, histories, comparisons):
        if not (len(comparisons) + 1) == len(histories):
            raise LinearTaxonomyException('Incompatible length lists of '
                                          'histories and comparisons')
        inter = self.interleave(histories, comparisons)
        for i in range(1, len(inter), 2):
            h0 = inter[i-1]
            comp = inter[i]
            h1 = inter[i+1]
            if not (h0.analyzedLevel == comp.lev1 and
                    comp.lev2 == h1.analyzedLevel):
                raise LinearTaxonomyException('Histories and comparisons do '
                                              'not overlap')

    def interleave(self, histories, comparisons):
        return [item for item in itertools.chain(
                    *itertools.zip_longest(histories, comparisons))
                    if item is not None]


# class LinearTaxonomyMultiSpecies(LinearTaxonomy):
#     def __init__(self, taxonomy, histories, species_histories):
#         super().__init__(taxonomy, histories, None)
#         top = histories[-1]
#         species_comparisons = [top.compare(h) for h in species_histories]
#         for i, h in enumerate(species_histories):
#             if not taxonomy.is_ancestor_of(top.analyzedLevel, h.analyzedLevel):
#                 raise LinearTaxonomyException('Species {} is not a descendant '
#                                               'of {}'.format(h.analyzedLevel,
#                                                              top.analyzedLevel))
#             node = TaxNode(h.analyzedLevel)
#             node.attach_fam_history(h)
#             node.attach_level_comparison_result(species_comparisons[i])
#             self.hierarchy[top.analyzedLevel].add_child(node)
#             node.add_parent(self.hierarchy[top.analyzedLevel])
#             self.hierarchy[node.name] = node
#             self.histories[node.name] = h


class XMLTaxonomy(Taxonomy):
    def __init__(self, filename):
        raise NotImplementedError("XML Taxonomies have not "
                                  "yet been implemented")


class TaxRangeOrthoXMLTaxonomy(Taxonomy):
    def __init__(self, parser):
        self.parser = parser
        self.extractAdjacencies()
        self.bloat_all()
        self.extractHierarchy()
        self.finialize_init()

    def _parseParentChildRelsR(self, grp):
        levels = None
        if self.parser.is_ortholog_group(grp):
            levels = [l.get('value') for l in grp.findall(
                './{{{ns0}}}property[@name="TaxRange"]'
                .format(**self.parser.ns))]
        directChildNodes = list(grp)
        children = [child for child in directChildNodes
                    if self.parser.is_evolutionary_node(child)]
        geneRefs = [node for node in directChildNodes
                    if OrthoXMLQuery.is_geneRef_node(node)]
        speciesOfGenes = {self.parser.mapGeneToSpecies(x.get('id'))
                          for x in geneRefs}

        # recursively process childreen nodes
        subLevs = speciesOfGenes
        for child in children:
            subLevs.update(self._parseParentChildRelsR(child))

        if levels is not None:
            for parent in levels:
                for child in subLevs:
                    self.adj.add((parent, child))
            subLevs = set(levels)
        return subLevs

    def extractAdjacencies(self):
        self.adj = set()
        for grp in self.parser.getToplevelGroups():
            self._parseParentChildRelsR(grp)

        self.nodes = set(itertools.chain(*self.adj))

    def bloat_all(self):
        """build transitive closure of all parent - child relations"""
        while(self.bloat()):
            pass

    def bloat(self):
        found = False
        for pair in itertools.product(self.nodes, repeat=2):
            first, second = pair
            for node in self.nodes:
                if ((pair not in self.adj) and ((first, node) in self.adj) and
                        ((node, second) in self.adj)):
                    found = True
                    self.adj.add(pair)
        return found

    def extractHierarchy(self):
        self.hierarchy = dict(zip(self.nodes, map(TaxNode, self.nodes)))
        for pair in itertools.product(self.nodes, repeat=2):
            if pair in self.adj:
                if self.good(pair):
                    first, second = pair
                    #print "%s,%s is good" % pair
                    self.hierarchy[first].add_child(self.hierarchy[second])
                    self.hierarchy[second].add_parent(self.hierarchy[first])
        noParentNodes = [z for z in self.nodes if self.hierarchy[z].up is None]
        if len(noParentNodes) != 1:
            raise TaxonomyInconsistencyError(
                "Warning: several/none TaxonomyNodes are roots: {}"
                .format(noParentNodes))
        self.root = noParentNodes[0]

    def good(self, pair):
        first, second = pair
        for node in self.nodes:
            if (first, node) in self.adj and (node, second) in self.adj:
                return False
        return True


class TaxNode(object):

    reg = re.compile(r'\W') # matches anything that's NOT a-z, A-Z, 0-9 or _

    def __init__(self, name, branch_length=None):
        self.name = name
        self.up = None
        self.down = list()
        self.history = None
        self.comparison = None
        self.branch_length = branch_length

    def __str__(self):
        if self.reg.search(self.name):
            label = '"{}"'.format(self.name)
        else:
            label = self.name
        branch_len_str = ":{}".format(self.branch_length) if self.branch_length else ""

        NHX = self._get_node_NHX() + self._get_edge_NHX()
        if self.is_leaf():
            return '{0}{1}{2}'.format(label, branch_len_str,
                            ('[&&NHX{0}]'.format(NHX) if NHX > '' else ''))
        subtree = ', '.join(str(ch) for ch in self.down)
        return '({0}){1}{2}{3}'.format(subtree, label, branch_len_str,
                            ('[&&NHX{0}]'.format(NHX) if NHX > '' else ''))

    def _get_node_NHX(self):
        return (':Genes={}'.format(len(self.history))
                if self.history else '')

    def _get_edge_NHX(self):
        if self.comparison:
            if not hasattr(self.comparison, 'summary'):
                self.comparison.summarise()

            n_ident     = self.comparison.summary['identical']
            n_dupl      = self.comparison.summary['duplicated']
            n_lost      = self.comparison.summary['lost']
            n_novel     = self.comparison.summary['novel']
            n_singleton = self.comparison.summary['singleton']

            NHX = (':Identical={0}'
                   ':Duplicated={1}'
                   ':Lost={2}'.format(n_ident,
                                      n_dupl,
                                      n_lost))
            if self.is_leaf():
                NHX += ':Novel=0:Singleton={}'.format(n_singleton)
            else:
                NHX += ':Novel={}:Singleton=0'.format(n_novel)

            return NHX
        return ''

    def add_child(self, c):
        if not c in self.down:
            self.down.append(c)

    def add_parent(self, p):
        if self.up is not None and self.up != p:
            raise TaxonomyInconsistencyError(
                "Level {} has several parents, at least two: {}, {}"
                .format(self.name, self.up.name, p.name))
        self.up = p

    def is_leaf(self):
        return len(self.down) == 0

    def is_inner(self):
        return not self.is_leaf()

    def is_root(self):
        return self.up is None

    def iter_preorder(self):
        """ Traverse the tree in preorder ordering: ancestors before
        descendents """
        yield self
        for child in self.down:
            for elem in child.iter_preorder():
                yield elem

    def iter_postorder(self):
        """ Traverse the tree in postorder ordering: descendents before
        ancestors """
        for child in self.down:
            for elem in child.iter_postorder():
                yield elem
        yield self

    def iter_levelorder(self):
        """ Traverse the tree in levelorder ordering: first the root, then
        nodes at distance==1 from the root, then distance==2, etc. """
        q = Queue()
        q.enqueue(self)
        while q:
            node = q.dequeue()
            yield node
            for child in node.down:
                q.enqueue(child)

    def iter_leaves(self):
        for elem in self.iter_preorder():
            if elem.is_leaf():
                yield elem

    def iter_inner_nodes(self):
        for elem in self.iter_preorder():
            if elem.is_inner():
                yield elem

    def attach_fam_history(self, history):
        self.history = history

    def attach_level_comparison_result(self, comparison):
        self.comparison = comparison


class NewickTaxonomy(Taxonomy):

    """ Create a taxonomy from a file or filehandle in newick format. The
    file should contain one tree (further trees are ignored). Only the
    tree topology is used - branch lengths and bootstrap support values
    are thown away. The leaf labels should match those in the orthoXML.
    Inner labels should match too, but for OMA XML will be automatically
    generated if auto_annotate == True """

    def __init__(self, fp):
        if isinstance(fp, basestring):
            if not os.path.exists(fp):
              raise Exception('File not found: {0}'.format(fp))
            fp = open(fp)
        self.lexer = NewickLexer(Streamer(fp))
        self.nodes = set()
        self.hierarchy = {}
        self.stack = []
        self.parse()
        self.finialize_init()

    def _get_label(self, tokens):
        """ Get the node data attributes 'label' and 'length'. Assumes these
        will be the next tokens in the stream. Throws ParseError if they are
        not. """
        label = next(self.lexer)
        if label.typ not in (tokens.LABEL, tokens.SUPPORT):
            raise ParseError(
                'Expected a label or a support value, found {0}'.format(
                    label))

        length = next(self.lexer)
        if length.typ != tokens.LENGTH:
            raise ParseError('Expected a length, found {0}'.format(
                length))

        return (label.val if label.typ == tokens.LABEL else None,
                length.val if length.typ == tokens.LENGTH else None)

    def annotate_from_orthoxml(self, xmlparser):
        """ Transfers internal node names from OMA orthoxml """
        def _annotate(self, node, levels_dict):
            if not node.down:
                self.nodes.add(node.name)
                self.hierarchy[node.name] = node
                return
            key = frozenset((n.name for n in node.iter_leaves()))
            node.name = levels_dict[key]
            self.nodes.add(node.name)
            self.hierarchy[node.name] = node
            for child in node.down:
                _annotate(self, child, levels_dict)

        levels = xmlparser.getLevels()
        levels_dict = dict((frozenset(x.split('/')), x) for x in levels)
        root_node = self.hierarchy[self.root]
        root_node.name = 'LUCA'
        self.hierarchy = {'LUCA': root_node}
        self.nodes = set('LUCA')
        self.root = 'LUCA'

        for child in root_node.down:
            _annotate(self, child, levels_dict)

    def populate(self, root):
        if len(root.down) == 0:
            self.hierarchy[root.name] = root
            self.nodes.add(root.name)
            return

        if not root.up:
            self.root = (root.name or 'LUCA')

        self.hierarchy[root.name] = root
        self.nodes.add(root.name)

        for child in root.down:
            self.populate(child)

    def parse(self):
        tmp_name = 1
        tokens = self.lexer.tokens

        for token in self.lexer:
            if token.typ == tokens.EOF:
                return

            elif token.typ == tokens.TREE:
                # invent a name for the node
                n = TaxNode(str(tmp_name))
                tmp_name += 1

                # push it onto the stack
                self.stack.append(n)

                # set as root
                self.root = n

            elif token.typ == tokens.SUBTREE:
                # invent a name and make a node
                n = TaxNode(str(tmp_name))
                tmp_name += 1

                # get parent
                p = self.stack[-1]
                p.add_child(n)
                n.add_parent(p)

                # push subtree onto the stack
                self.stack.append(n)

            elif token.typ == tokens.LEAF:
                label, branch_len = self._get_label(tokens)
                self.nodes.add(label)
                l = TaxNode(label, branch_len)

                # get parent from stack
                p = self.stack[-1]
                p.add_child(l)
                l.add_parent(p)

            elif token.typ == tokens.ENDSUB:
                label, branch_len = self._get_label(tokens)

                # retrieve node from stack
                subtree = self.stack.pop()

                # update node name
                if isinstance(label, str):
                    subtree.name = label
                if isinstance(branch_len, numbers.Number):
                    subtree.branch_length = branch_len

            elif token.typ == tokens.ENDTREE:  # trigger for tree-finalising functions
                self.populate(self.root)
                del self.lexer
                return

            elif token.typ in (tokens.LABEL, tokens.LENGTH, tokens.SUPPORT):
                raise ParseError('Unexpected token in stream: {0}'.format(token))

            else:
                raise ParseError('Not sure what happened')



from numbers import Integral, Real
from six import string_types
from abc import ABCMeta, abstractproperty
from dendropy import Tree


class Option(object):
    """Abstract base class for an option.

    Options provide an interface between the wrapper and the
    concrete command line option of the wrapped program."""
    __metaclass__ = ABCMeta

    def __init__(self, name, default=None, active=False):
        self._name = name
        self.set_value(default)
        self.active = active

    def __repr__(self):
        return '{}({}={}) <{}>'.format(self.__class__.__name__, self.name, self.get_value(), 'on' if self.active else 'off')

    def __str__(self):
        return (' '.join([self._name, str(self.get_value())]) if self.active else '')

    @property
    def active(self):
        return self._active

    @active.setter
    def active(self, val):
        self._active = True if val else False

    @property
    def name(self):
        return self._name

    def set_value(self, value):
        self._value = value
        if value is not None:
            self.active = True

    def get_value(self):
        return self._value

    def set_and_activate(self, value):
        self.set_value(value)
        self.active = True

    def status(self):
        return 'Name: {}\nValue: {}\nActive: {}\nStr: {}'.format(self.name,
                                                                 self.get_value(),
                                                                 self.active,
                                                                 str(self) or "''")


class ValueOption(Option):
    __metaclass__ = ABCMeta


class TypedValueOption(ValueOption):
    """A TypedValueOption is an option that only accepts options of a given type.

    This abstract class provides the functionality to check the type
    of a passed value and raises an ValueError if it doesn't match
    the expected type.

    A TypedValueOption must overwrite the abstract property _type.
    """

    __metaclass__ = ABCMeta

    @abstractproperty
    def _type(self):
        pass

    def set_value(self, value):
        if isinstance(value, self._type):
            self._value = value
            self.active = True

        else:
            raise ValueError('Value should be of type {}'.format(self.type))


### Concrete classes from here on

class IntegerOption(TypedValueOption):
    """option to hold an integer value"""
    @property
    def _type(self):
        return Integral


class FloatOption(TypedValueOption):
    """Option to hold a real number value"""

    @property
    def _type(self):
        return Real

    def get_value(self):
        return float(self._value)


class StringOption(TypedValueOption):
    """Opion to hold a string value"""

    def __init__(self, name, value=None, active=False):
        if value is None:
            value = str()
        super(StringOption, self).__init__(name, value, active)

    @property
    def _type(self):
        return string_types


class FlagOption(TypedValueOption):
    """Option to hold a boolean flag value, i.e. True or False"""
    @property
    def _type(self):
        return bool

    def __str__(self):
        return (self._name if self.active and self.get_value() else '')


class TreeInputOption(TypedValueOption):
    """Option to hold a phylogenetic tree argument.

    As of now, Trees are represented as :class:`dendropy.Tree` objects."""

    @property
    def _type(self):
        return Tree


class MultiOption(Option):
    """Option to hold a list"""

    @property
    def _type(self):
        return list

    def __str__(self):
        listopts = self.get_value()
        if listopts is None: return ''
        strings = []
        for item in listopts:
            item_string = ' '.join([self._name, str(item)]) if self.active else ''
            if item_string > '':
                strings.append(item_string)

        return ' '.join(strings)


class OptionSet(object):
    """Option to hold a set of key-value pairs."""
    def __init__(self, options):
        if isinstance(options, (list, tuple)):
            self.options = {opt.name: opt for opt in options}
        elif isinstance(options, dict):
            self.options = options
        else:
            raise ValueError('Expected a list, tuple or dict of options, not {}'.format(type(options)))

    def __str__(self):
        strings = []
        for name, option in self.options.items():
            option_string = str(option)
            if option_string > '':
                strings.append(option_string)

        return ' '.join(strings)

    def __getitem__(self, item):
        return self.options[item]

    def list(self):
        return [(name, option) for (name, option) in self.options.items()]



import logging
import dendropy as dpy
from pyparsing import Suppress, SkipTo, Word, Regex, Literal, OneOrMore, Group, LineEnd, CharsNotIn, nums, alphanums, \
    ParseException


logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())


FLOAT = Word(nums + '.-').setParseAction(lambda x: float(x[0]))
INT = Word(nums).setParseAction(lambda x: int(x[0]))
WORD = Word(alphanums + '_')
SPACEDWORD = Word(alphanums+' _')


class ProtTestParser(object):
    """
    Simple prottest result parser.
    """

    def __init__(self):
        self.MODEL = Regex(r'Best model according to\s+')
        # These are all the models that are possible to be tested using phyml
        self.model = OneOrMore(Group(Suppress(SkipTo(self.MODEL)) + Suppress(self.MODEL) + WORD + Suppress(":") + WORD))

    def parse(self, s):
        model = None
        try:
            model = self.model.parseString(s).asList()
        except ParseException as err:
            logger.error(err)

        return model

    def to_dict(self, stats_filename):
        result = {}
        model = self.parse(stats_filename)
        try:
            for mg in model:
                result[mg[0]] = mg[1]
        except IOError as err:
            logger.error(err)
            return

        return result







#!/usr/bin/env python
'''
    Utilities for zoo files.
'''
from io import BytesIO
import bz2
import gzip
import os
import logging
logger = logging.getLogger(__name__)


# File opening. This is based on the example on SO here:
# http://stackoverflow.com/a/26986344
fmagic = {b'\x1f\x8b\x08': gzip.open,
          b'\x42\x5a\x68': bz2.BZ2File}


def auto_open(fn, *args, **kwargs):
    """function to open regular or compressed files for read / write.

    This function opens files based on their "magic bytes". Supports bz2
    and gzip. If it finds neither of these, presumption is it is a
    standard, uncompressed file.

    Example::

        with auto_open("/path/to/file/maybe/compressed", mode="rb") as fh:
            fh.read()

        with auto_open("/tmp/test.txt.gz", mode="wb") as fh:
            fh.write("my big testfile")

    :param fn: either a string of an existing or new file path, or
        a BytesIO handle
    :param \*\*kwargs: additional arguments that are understood by the
        underlying open handler
    :returns: a file handler
    """
    if isinstance(fn, BytesIO):
        return fn

    if os.path.isfile(fn) and os.stat(fn).st_size > 0:
        with open(fn, 'rb') as fp:
            fs = fp.read(max([len(x) for x in fmagic]))
        for (magic, _open) in fmagic.items():
            if fs.startswith(magic):
                return _open(fn, *args, **kwargs)
    else:
        if fn.endswith('gz'):
            return gzip.open(fn, *args, **kwargs)
        elif fn.endswith('bz2'):
            return bz2.BZ2File(fn, *args, **kwargs)

    return open(fn, *args, **kwargs)


class LazyProperty(object):
    """Decorator to evaluate a property only on access.

    Compute the attribute value and caches it in the instance.
    Python Cookbook (Denis Otkidach) http://stackoverflow.com/users/168352/denis-otkidach
    This decorator allows you to create a property which can be computed once and
    accessed many times.

    Example::

        class Circle:
            def __init__(self, radius):
                self.radius = radius

            @LazyProperty
            def area(self):
                print("computing area")
                return 3.14 * self.radius ** 2

        >>> c = Circle(4)
        >>> c.area
        computing area
        50.24
        >>> c.area
        50.24

    You can see that the property method is only executed once.
    """

    def __init__(self, method, name=None):
        # record the unbound-method and the name
        self.method = method
        self.name = name or method.__name__
        self.__doc__ = method.__doc__

    def __get__(self, inst, cls):
        if inst is None:
            return self
        # compute, cache and return the instance's attribute value
        result = self.method(inst)
        # setattr redefines the instance's attribute so this doesn't get called again
        setattr(inst, self.name, result)
        return result


def unique(seq):
    """Return the elements of a list uniquely while preserving the order

    :param list seq: a list of hashable elements
    :returns: new list with first occurence of elements of seq"""
    seen = set()
    return [x for x in seq if x not in seen and not seen.add(x)]






import collections
import csv
import networkx as nx

from Bio import SeqIO
import pickle
from os import listdir
import os

from ._utils_subhog import logger_hog
from .zoo.unionfind import UnionFind
from . import _config
HOGMapData = collections.namedtuple("HOGMapData", ("hogid", "score", "seqlen", "subfamily_medianseqlen"))

import collections

"""UnionFind.py

Union-find data structure. Based on Josiah Carlson's code,
http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/215912
with significant additional changes by D. Eppstein and
Adrian Altenhoff.
"""


class UnionFind(object):
    """Union-find data structure.

    Each unionFind instance X maintains a family of disjoint sets of
    hashable objects, supporting the following two methods:

    - X[item] returns a name for the set containing the given item.
      Each set is named by an arbitrarily-chosen one of its members; as
      long as the set remains unchanged it will keep the same name. If
      the item is not yet part of a set in X, a new singleton set is
      created for it.

    - X.union(item1, item2, ...) merges the sets containing each item
      into a single larger set.  If any item is not yet part of a set
      in X, it is added to X as one of the members of the merged set.
    """

    def __init__(self, elements=None):
        """Create a new union-find structure.

        If elements is not None, the structure gets initialized
        with each element as a singleton component.

        :param elements: an iterable to initialize the structure.
        """

        self.weights = {}
        self.parents = {}
        if elements is not None:
            for elem in iter(elements):
                self.parents[elem] = elem
                self.weights[elem] = 1

    def __getitem__(self, obj):
        """return the name of set which contains obj.

        :param obj: the query object

        :SeeAlso: :meth:`find`"""
        return self.find(obj)

    def find(self, obj):
        """Find and return the name of the set containing the obj.

        If the object is not found in any set, a new singleton set
        is created that holds only this object until it is further merged."""

        # check for previously unknown obj. If unknown, add it
        # as a new cluster
        if obj not in self.parents:
            self.parents[obj] = obj
            self.weights[obj] = 1
            return obj

        # find path of objects leading to the root
        path = [obj]
        root = self.parents[obj]
        while root != path[-1]:
            path.append(root)
            root = self.parents[root]

        # compress the path and return
        for ancestor in path:
            self.parents[ancestor] = root
        return root

    def remove(self, obj):
        """Remove an object from the sets.

        Removes an object entirly from the datastructure. The
        containing set will shrink by this one element.

        :Note: If one tries to accessed it afterwards using
            :meth:`find`, it will be created newly and put as a
            singleton.
        """
        if obj not in self.parents:
            return
        comp = self.find(obj)
        self.weights[comp] -= 1
        self.parents.pop(obj)

    def __iter__(self):
        """Iterate through all items ever found or unioned by this structure."""
        return iter(self.parents)

    def union(self, *objects):
        """Find the sets containing the objects and merge them.

        any number of objects can be passed to this method and
        all of them will be merged into one set containing at
        least these objects.

        :param objects: the objects to be merged. they have to be all
            hashable. If they haven't been initialy added to the UnionFind
            datastructre at instantiation time, they are added at this point
            in time.
        """
        roots = [self[x] for x in objects]
        heaviest = max([(self.weights[r], r) for r in roots], key=lambda x: x[0])[1]
        for r in roots:
            if r != heaviest:
                self.weights[heaviest] += self.weights[r]
                self.parents[r] = heaviest

    def get_components(self):
        """return a list of sets corresponding to the connected
        components of the structure."""
        comp_dict = collections.defaultdict(set)
        for elem in iter(self):
            comp_dict[self[elem]].add(elem)
        comp = list(comp_dict.values())
        return comp


def parse_proteomes(folder=None):  # list_oma_species
    """
    parsing fasta files of proteins located in /proteome/
    using Bio.SeqIO.parse
    Each fasta file is for one species.  The file name is the species name.
    All proteoems should be with the same extension.
    output: prot_recs_lists: a dic with key as species name and  its value list of Biopython record of species.
    """
    if folder is None:
        folder = "./proteome"

    fasta_format_keep = ""
    project_files = listdir(folder)
    logger_hog.debug("using '%s' as proteome folder, found %d files", folder, len(project_files))
    species_names = []  # query/input species name based on the file name
    for file in project_files:
        species_name, ext = file.rsplit('.', 1)
        logger_hog.debug("%s: %s | %s", file, species_name, ext)
        if ext in ("fa", "fasta"):
            species_names.append(species_name)
            fasta_format_keep = ext  # last one is stored either fa or fasta

    # todo accept all fasta formats in the input prtoeome folder, fasta, fa, fna, ..
    prot_recs_lists = {} # key: species name, value is a dic of query protein Biopython records. # 'MYCGE': [SeqRecord(seq=Seq('MDFDK
    for species_name in species_names:
        prot_address = os.path.join(folder, species_name + "." + fasta_format_keep)
        prots_record = list(SeqIO.parse(prot_address, "fasta"))
        prot_recs_lists[species_name] = prots_record

    logger_hog.info("There are %d species in the proteome folder.", len(species_names))
    return species_names, prot_recs_lists, fasta_format_keep


def add_species_name_prot_id(species_names, prot_recs_lists):
    """
    adding the name of species to each protein record
        - based on file name
    adding protein idx number, integer needed by xml format
    output:  prot_recs_all =  {'MYCGE': {'sp|P47500|RF1_MYCGE||MYCGE||1000000001': SeqRecord(seq=
    """
    prot_idx_name_pickle_file = "./gene_id_dic_xml.pickle"
    start_num_prot = int(1e9)
    start_num_prot_per_sp = int(1e6) #
    prot_recs_all = {} # {'MYCGE': {'sp|P47500|RF1_MYCGE||MYCGE||1000000001': SeqRecord(seq=
    prot_idx_name = {} # {'MYCGE': [(1000000001, 'sp|P47500|RF1_MYCGE'),(1000000002, 'sp|P13927|EFTU_MYCGE'),
    species_idx = -1
    for species_name, prot_recs_list in prot_recs_lists.items():
        species_idx += 1
        prot_idx = start_num_prot + species_idx * start_num_prot_per_sp
        prot_recs_all[species_name]={}
        prot_idx_name[species_name] = []
        for prot_rec in prot_recs_list:
            prot_idx+=1
            prot_name= prot_rec.id

            if len(prot_name) > 230:
                logger_hog.info("We are truncating the prot name as it may be problamatic for mafft, " + str(prot_name))
                prot_name = prot_name[:230]

            # todo, this could be a dic
            prot_idx_name[species_name].append((prot_idx, prot_name))

            prot_name_new = prot_name+ "||"+species_name+"||"+str(prot_idx) # orthoxml file needs an integer as
            prot_rec.id = prot_name_new
            prot_recs_all[species_name][prot_name] = prot_rec

    with open(prot_idx_name_pickle_file, 'wb') as handle:
        pickle.dump(prot_idx_name, handle, protocol=pickle.HIGHEST_PROTOCOL)

    return prot_recs_all


def parse_hogmap_omamer(species_names, fasta_format_keep,  folder=None):
    """
     function for parsing output of omamer (hogmap files) located in /hogmap/
    Each hogmap file correspond to one fasta file of species, with the same name.
    Note that some records of fasta may removed in hogmap, probably becuase of being so short.
    hogmap file example:
    # qseqid	hogid	hoglevel	family_p	family_count	family_normcount	subfamily_score	subfamily_count	qseqlen	subfamily_medianseqlen	qseq_overlap
    # sp|O66907|ATPA_AQUAE	HOG:C0886513.1b	Eukaryota	696.5519485850672	187	0.37302594463771466	0.2643067275644273	120	504	551	0.8230616302186878
    output is dic of dic for all species:
    """
    if folder is None:
        folder = "./hogmap"

    hogmaps = {}
    unmapped = collections.defaultdict(list)
    for species_name in species_names:
        hogmap_address = os.path.join(folder, species_name + "." + fasta_format_keep + ".hogmap")
        cur_species_hogmap = collections.defaultdict(list)
        with open(hogmap_address, 'rt') as hogmap_file:
            reader = csv.DictReader((line for line in hogmap_file if not line.startswith('!')),
                                    dialect="excel-tab")
            for row in reader:
                if row['hogid'] == "N/A":
                    unmapped[species_name].append(row['qseqid'])
                else:
                    cur_species_hogmap[row['qseqid']].append(
                        HOGMapData(row['hogid'], row['family_p'], row['qseqlen'], row['subfamily_medianseqlen']))
        hogmaps[species_name] = cur_species_hogmap
        logger_hog.info("hogmap %s: %d proteins mapped to %d hogs, %d not mapped",
                        species_name, len(cur_species_hogmap), sum(len(x) for x in cur_species_hogmap.values()),
                        len(unmapped[species_name]))

    logger_hog.info("There are %d species in the hogmap folder.", len(hogmaps))
    return hogmaps, unmapped


def group_prots_roothogs(hogmaps):
    """
    function for finding those proteins that are mapped to the same rootHOG.
    output: rhogs_dic
    """

    rhogs_prots = {}
    for species_name, prots_map in hogmaps.items():
        # prot_recs = prot_recs_all[species_name]

        for prot_id, prot_map in prots_map.items():
            # omamer output is sorted based on normcount (not family-p). but that's intended by omamer developers
            # this helps me in other functions like handle_singleton in this
            #  this should be commented
            # if len(prot_map)>1:
            #     scores = [float(i[1]) for i in prot_map] # (hogid,score,seqlen,subfamily_medianseqlen)
            #     rhogids =[i[0] for i in prot_map]
            #     # select hog with the best score
            #     max_index = scores.index(max(scores))
            #     hogid =rhogids[max_index]
            # else:
            #     hogid = prot_map[0][0]
            hogid = prot_map[0][0]
            rhogid = hogid.split(".")[0].split(":")[1]

            if rhogid in rhogs_prots:
                rhogs_prots[rhogid].append((species_name, prot_id))
            else:
                rhogs_prots[rhogid] = [(species_name, prot_id)]
    logger_hog.info("There are " + str(len(rhogs_prots)) + " rootHOGs.")
    # from old code, to check later: Keep prot seq in hog class. We can write species idx and prot idx to  improve speed of code for omamer tresholidng
    return rhogs_prots


def handle_singleton(rhogs_prots,hogmaps):
    query_singleton_rhog = []  # they are the only one to hit to this HOG
    for rhog, prot_sp_list in rhogs_prots.items():  # 'C0602129': [('GORGO', 'GORGO00025'), ('GORGO', 'GORGO00026')],
        if len(prot_sp_list) == 1:
            query_singleton_rhog += prot_sp_list
    logger_hog.debug(
        "There are " + str(len(query_singleton_rhog)) + " singleton. The hog on which only one query proteins mapped.")

    # adding singleton to already rhogs with >1 members
    query_singleton_rhog_solved = []
    for (species, prot) in query_singleton_rhog:
        prot_maps = hogmaps[species][prot]
        if len(prot_maps) > 1:
            #prot_maps2 = prot_maps
            # omamer output is sorted based on normcount. but that's intended by omamer developers
            # family_p	family_count	family_normcount
            scores = [float(i[1]) for i in prot_maps]  # (hogid,score,seqlen,subfamily_medianseqlen)
            hogids = [i[0] for i in prot_maps]
            rhogid0 = hogids[0].split(".")[0].split(":")[1]  # 'HOG:D0903929.1a'
            for idx, hogid in enumerate(hogids[1:]):  # the 0-index is already checked and is a singleton
                rhogid = hogid.split(".")[0].split(":")[1]
                if rhogid in rhogs_prots:
                    if len(rhogs_prots[rhogid]) > 1 and scores[idx+1] >  _config.threshod_f_score_merging: # todo  check
                        rhogs_prots[rhogid].append((species, prot))
                        del rhogs_prots[rhogid0]
                        query_singleton_rhog_solved.append((species, prot))
                        break
    logger_hog.debug("We add " + str(
        len(query_singleton_rhog_solved)) + " proteins/singleton to another rHOG based on omamer multi-hit.")

    query_singleton_rhog_solved_set = set(query_singleton_rhog_solved)
    query_singleton_rhog_remained = [i for i in query_singleton_rhog if i not in query_singleton_rhog_solved_set]
    logger_hog.debug("However, " + str(len(query_singleton_rhog_remained)) + " proteins/singletons are remained.")

    # try to create group from multi-hits of remained singleton
    dic_singlton_remained = {}
    for (species, prot) in query_singleton_rhog_remained:
        prot_maps = hogmaps[species][prot]
        scores = [float(i[1]) for i in prot_maps]  # (hogid,score,seqlen,subfamily_medianseqlen)
        hogids = [i[0] for i in prot_maps]
        for idx, hogid in enumerate(hogids[1:]):  # the 0-index is already checked and is a singleton
        #for hogid in hogids[1:]:
            if scores[idx+1] >  _config.threshod_f_score_merging :
                rhogid = hogid.split(".")[0].split(":")[1]
                if rhogid in dic_singlton_remained:
                    dic_singlton_remained[rhogid].append((species, prot))
                else:
                    dic_singlton_remained[rhogid] = [(species, prot)]
    logger_hog.debug("These are associated to " + str(len(dic_singlton_remained)) + " HOGs considering all multi-hits.")

    prots_rhogs_dic = {}
    for rhog, sp_prots in rhogs_prots.items():
        if len(sp_prots) == 1:
            prots_rhogs_dic[sp_prots[0]] = rhog

    count_new_rhogids = 0
    count_updated_rhogids = 0
    already_grouped = []
    for rhogid, spec_prot_list in dic_singlton_remained.items():
        if len(set(spec_prot_list)) > 1:
            already_grouped_set = set(already_grouped)
            val = sum([1 for i in spec_prot_list if i in already_grouped_set])

            if val == 0:  # the prot has not yet grouped
                if rhogid in rhogs_prots:
                    count_updated_rhogids += 1
                else:
                    count_new_rhogids += 1
                #remove singleton from rhogs_prots
                for spec_prot in spec_prot_list:
                    rhogid_ii = prots_rhogs_dic[spec_prot]
                    del rhogs_prots[rhogid_ii]

                rhogs_prots[rhogid] = spec_prot_list
                already_grouped += spec_prot_list
    logger_hog.debug("We updated " + str(count_updated_rhogids) + " and created  " + str(
        count_new_rhogids) + " new HOGs  which are not singleton anymore.")

    counter_rhog_singlton_final = 0
    counter_notsingl_final = 0
    for rhogid, spec_prot_list in rhogs_prots.items():
        if len(set(spec_prot_list)) > 1:
            counter_notsingl_final += 1
        else:
            counter_rhog_singlton_final += 1

    logger_hog.debug("Now, we have " + str(counter_notsingl_final) + " rootHOGs with >1 proteins and  and " + str(
        counter_rhog_singlton_final) + " singleton rootHOGs ")

    return rhogs_prots


def filter_big_roothogs(hogmaps, rhogs_prots):


    prots_list_big = []  # as backup
    rhogids_big = []
    for rhogid, sp_prot_list in rhogs_prots.items():
        if len(sp_prot_list) > _config.big_rhog_size:
            prots_list_big.append(sp_prot_list)  # (species_name,prot_id)
            rhogids_big.append(rhogid)
            logger_hog.info(
                "a big rootHOG was found " + rhogid + " with " + str(len(sp_prot_list)) + " query proteins.")
    logger_hog.info("There are " + str(len(rhogids_big)) + " big rootHOGs.")

    for rhogid in rhogids_big:

        sp_prot_list = rhogs_prots[rhogid]
        sp_prot_list_filt = []
        hogids = []

        # removing proteins with low omamer score from big rootHOG
        for species_name, prot_id in sp_prot_list:
            prot_maps = hogmaps[species_name][prot_id]
            # if len(prot_maps) > 1:  # probably for big rootHOG, there won't be multi-hits
            #     scores = [float(i[1]) for i in prot_maps]  # (hogid,score,seqlen,subfamily_medianseqlen)
            #     hogids = [i[0] for i in prot_maps]
            #     max_score = max(scores)
            #     max_index = scores.index(max_score)
            #     hogid = hogids[max_index]
            # else:
            #     hogid = prot_maps[0][0]
            #     max_score = float(prot_maps[0][1])
            # todo: highest normcount or pavalue , default of omamer ?
            hogid = prot_maps[0][0]
            max_score = float(prot_maps[0][1])

            if max_score > _config.omamer_family_threshold:
                sp_prot_list_filt.append((species_name, prot_id))
                hogids.append(hogid)

        logger_hog.info("For big rootHOG " + rhogid + ", " + str(
            len(sp_prot_list_filt)) + " proteins left after filtering with threshold " + str(_config.omamer_family_threshold))

        if len(sp_prot_list_filt) < _config.big_rhog_size:
            sp_prot_list_filt2 = sp_prot_list_filt

        elif  _config.filter_nonchild_rootHOG :  # removing proteins that are mapped to rootHOG (= HOGC123 not the levelsHOGC123.1a) from big rootHOG
            #hogids2 = []
            sp_prot_list_filt2 = []
            for prot_idx, sp_prot in enumerate(sp_prot_list_filt): #[('UP000192223_224129', 'tr|A0A1W4WAU6|A0A1W4WAU6_AGRPL'), ('UP000192223_224129', 'tr|A0A1W4WU99|A0A1W4WU99_AGRPL'),
                species , protein = sp_prot
                prot_maps =hogmaps[species][protein]
                hogid = prot_maps[0][0]
                if len(hogid.split(".")) > 1: # if mapped to subHOG (not to the rootHOG)
                    sp_prot_list_filt2.append(sp_prot)
                    #hogids2.append(hogid)
        else:
            sp_prot_list_filt2 = sp_prot_list_filt

        logger_hog.info("For big rootHOG " + rhogid + ", " + str(
            len(sp_prot_list_filt2)) + " proteins left after removing non-child subhogs if activate: "+str(_config.filter_nonchild_rootHOG))
        if len(sp_prot_list_filt2):
            rhogs_prots[rhogid] = sp_prot_list_filt2
        else:
            del rhogs_prots[rhogid]

    return rhogs_prots


def write_rhog(rhogs_prot_records, prot_recs_all, address_rhogs_folder, min_rhog_size=1):
    # max_rhog_size =1e12
    #address_rhogs_folder = folder + "rhog"
    logger_hog.info("Writing Sequences of roothogs are fasta file in " + address_rhogs_folder)
    if not os.path.exists(address_rhogs_folder):
        os.mkdir(address_rhogs_folder)


    rhogid_written_list = []
    for rhogid, rhog_prots in rhogs_prot_records.items():
        rhog_recs = []
        for (species_name, prot_name) in rhog_prots:
            prot_rec = prot_recs_all[species_name][prot_name]
            rhog_recs.append(prot_rec)

        if min_rhog_size <= len(rhog_recs):  # <= max_rhog_size:
            # todo add the release id   to file names  rhogids_list[:2] > ['HOG:C0884658', 'HOG:C0709155']
            SeqIO.write(rhog_recs, address_rhogs_folder + "/HOG_" + rhogid + ".fa", "fasta")
            rhogid_written_list.append(rhogid)
        else:
            for prot1 in rhog_recs:
                a=1
                #logger_hog.debug("we are removing due to omamer signleton hog |*|" + str(prot1.id))

    logger_hog.info("Writing Sequences of " + str(len(rhogid_written_list)) + " roothogs finished.")

    return rhogid_written_list


def find_rhog_candidate_pairs(hogmaps, rhogs_prots):

    pair_rhogs_count = {}
    rhogs_size = {}
    for rhog, prt_prot_maps in hogmaps.items():
        for prot, prot_maps in prt_prot_maps.items():
            # [('HOG:D0017631.5a', '1771.7328874713658', '253', '234'), ('HOG:D0863448', '163.60700392437903', '253', '244'),
            rhogids = []
            scores = []
            for prot_map in prot_maps:
                hog, score = prot_map[:2]
                if float(score) > _config.threshod_f_score_merging:
                    rhogid = hog.split(".")[0].split(":")[1]
                    if not rhogid in rhogs_size:
                        rhogs_size[rhogid]  = 0
                    rhogs_size[rhogid] += 1
                    rhogids.append(rhogid)
                    #rhogids.append(rhogid)
                    scores.append(float(score))

            for ii in range(len(rhogids)):
                for jj in range(ii + 1, len(rhogids)):
                    hogi, hogj = rhogids[ii], rhogids[jj]
                    # Below Yannis made it so the pair of HOG is ordered in alphabetic order, it makes it so (HOG:D0017631,HOG:D0863448) is considered the same as (HOG:D0863448,HOG:D0017631)
                    pair = tuple(sorted((hogi, hogj)))
                    if pair[0]==hogi:
                        score_i, score_j = scores[ii], scores[jj]
                    else:
                        score_j, score_i = scores[ii], scores[jj]

                    if pair in pair_rhogs_count:
                        pair_rhogs_count[pair][0].append(score_i)  # += 1
                        pair_rhogs_count[pair][1].append(score_j)  # += 1
                    else:
                        pair_rhogs_count[pair] = [[score_i], [score_j]]

                    # score_i, score_j = scores[ii], scores[jj]
                    # if (hogi, hogj) in pair_rhogs_count:
                    #     pair_rhogs_count[(hogi, hogj)][0].append(score_i)  # += 1
                    #     pair_rhogs_count[(hogi, hogj)][1].append(score_j)  # += 1
                    # else:
                    #     pair_rhogs_count[(hogi, hogj)] = [[score_i], [score_j]]
                    #     pair_rhogs_count[(hogi, hogj)] += 1
                    # else:
                    #     pair_rhogs_count[(hogi, hogj)] = 1

    print(len(pair_rhogs_count))
    logger_hog.debug("There are " + str(len(pair_rhogs_count)) + " pairs of rhogs.")

    #rhogs_size = {}
    #for rhog, list_prot in rhogs_prots.items():
    #    rhogs_size[rhog] = len(list_prot)

    candidates_pair = []
    # dic_pair_ratio = {}
    for (hogi, hogj), score_shared_lists in pair_rhogs_count.items():
        count_shared = len(score_shared_lists[0])
        list_lowest =[min(score_shared_lists[0][ik],score_shared_lists[1][ik]) for ik in range(len(score_shared_lists[0]))]
        mean_scores = sum(score_shared_lists[0]+score_shared_lists[1])/2*len(score_shared_lists[0])
        if hogi in rhogs_size and hogj in rhogs_size:  # during previous functions, we might
            ratioMax = count_shared / max(rhogs_size[hogi], rhogs_size[hogj])
            ratioMin = count_shared / min(rhogs_size[hogi], rhogs_size[hogj])
            #mean_scores > _config.mergHOG_mean_thresh or
            if ( max(rhogs_size[hogi], rhogs_size[hogj]) <  _config.big_rhog_size / 6 and ((ratioMax > _config.mergHOG_ratioMax_thresh or ratioMin >  _config.mergHOG_ratioMin_thresh ) and count_shared > _config.mergHOG_shared_thresh ) )\
                    and rhogs_size[hogi] < _config.big_rhog_size / 6 and  rhogs_size[hogj] < _config.big_rhog_size / 6:
                if rhogs_size[hogi] >= rhogs_size[hogj]:
                    candidates_pair.append((hogi, hogj))  # bigger first
                else:
                    candidates_pair.append((hogj, hogi))
                # print(hogi,"(",rhogs_size[hogi],")",hogj,"(",rhogs_size[hogj],")", count_shared,round(ratioMax,2),round(ratioMin,2))
            # D0651051 ( 59 ) D0658569 ( 180 ) 62 0.34 1.05 # todo why bigger than 1?
            # D0646495 ( 2 ) D0631227 ( 14 ) 25 1.79 12.5

    logger_hog.debug("There are " + str(len(candidates_pair)) + " candidate pairs of rhogs for merging.")
    print(len(candidates_pair))

    return candidates_pair

def cluster_rhogs(candidates_pair):
    cluster_rhogs_obj = UnionFind()
    for pair in candidates_pair:
        cluster_rhogs_obj.union(pair[0], pair[1])
    cluster_rhogs_sets = cluster_rhogs_obj.get_components()

    cluster_rhogs_list = [ ]
    for cl in cluster_rhogs_sets:
        cluster_rhogs_list.append(list(cl))

    return cluster_rhogs_list
#
# def cluster_rhogs_old(candidates_pair): # doesnt work correctly
#     # init
#     all_hog_raw = []
#     for pair in candidates_pair:
#         all_hog_raw.append(pair[0])
#         all_hog_raw.append(pair[1])
#     all_hog = list(set(all_hog_raw))
#
#     allcc = []  # connected compoenets
#     for hog in all_hog:
#         allcc.append([hog])
#
#     dic_where = {}
#     for idx, cc in enumerate(allcc):
#         dic_where[cc[0]] = idx  # in the beginning, each inner list has only on element, a unique hog
#
#     # print(len(all_hog_raw),len(all_hog),len(allcc),allcc[:2])
#     logger_hog.debug("There are " + str(len(all_hog)) + " all_hog.")
#
#     # print(dic_where)
#     for pair in candidates_pair:
#         # print(pair)
#         idx_0 = dic_where[pair[0]]
#         idx_1 = dic_where[pair[1]] # this will be problametic for sequence of pairs
#
#         dic_where[pair[1]] = idx_0
#
#         # print(idx_0)
#         allcc[idx_0] += allcc[idx_1]
#         allcc[idx_1] = []
#
#         # print(dic_where,allcc)
#
#     cluster_rhogs_list = [i for i in allcc if len(i) > 1]
#     # print(cluster_rhogs_list)
#     logger_hog.debug("There are " + str(len(cluster_rhogs_list)) + " cluster_rhogs.")
#
#     # allcc_cleaned_len = [len(i) for i in cluster_rhogs]
#     # len(allcc),len(cluster_rhogs),np.sum(allcc_cleaned_len)
#
#     return cluster_rhogs_list



def merge_rhogs2(hogmaps, rhogs_prots):
    logger_hog.debug("started merging  ")

    candidates_pair = find_rhog_candidate_pairs(hogmaps, rhogs_prots)

    print("There are " + str(len(candidates_pair)) + " candidate pairs of rhogs for merging.")
    cluster_rhogs_list = cluster_rhogs(candidates_pair)

    print("There are " + str(len(cluster_rhogs_list)) + " clusters.")

    cluster_rhogs_list = cluster_rhogs_nx(cluster_rhogs_list, candidates_pair)

    print("There are " + str(len(cluster_rhogs_list)) + " selected clusters.")

    #cluster_rhogs_list = cluster_rhogs(candidates_pair)



    num_rhog_g1 = 0
    for rhg, list_pr in rhogs_prots.items():
        if len(list_pr) > 1:
            num_rhog_g1 += 1

    logger_hog.debug("There are " + str(num_rhog_g1) + " rhogs (size>1) before merging.")
    print(len(rhogs_prots))

    file_out_merge =  open("merging_rhogs.txt","w")
    file_out_merge.write("#first column is the host hog, the rest will be merged here.\n")
    for cluster in cluster_rhogs_list:
        cluster_host = min(cluster) # write the host rhog as the first element
        file_out_merge.write(cluster_host + "\t")
        for cluster_i in cluster:
            if cluster_i != cluster_host:
                file_out_merge.write(cluster_i+"\t")
        file_out_merge.write("\n")
    file_out_merge.close()

    counter_merged_prots=0
    newhost= ""
    for cluster in cluster_rhogs_list:
        if len(cluster)>30:
            print(len(cluster))
        #prots = [rhogs_prots[hog] for hog in cluster]
         #  cluster[0] #
        found = False
        while not found and len(cluster)>0:
            host_hog = min(cluster)
            if host_hog in rhogs_prots:
                found = True
                rhogs_host_size = len(rhogs_prots[host_hog])
            else:
                # logger print rhog
                cluster.remove(host_hog)

        all_prots = []
        all_prots2 =[]
        for hog in cluster:
            # todo check, create smallers clusters
            if len(all_prots) < _config.big_rhog_size:
                if hog in rhogs_prots:
                    all_prots += rhogs_prots[hog]
                    del rhogs_prots[hog]
            else:
                if newhost:
                    if len(all_prots2) < _config.big_rhog_size:
                        if hog in rhogs_prots:
                            all_prots2 += rhogs_prots[hog]
                            del rhogs_prots[hog]
                    else:
                        print("cluster is very big "+str(len(cluster)))
                else:
                    newhost = hog

        if newhost:
            all_prots_uniq2 = list(set(all_prots2))  # there might be repeated prots
            rhogs_prots[newhost] = all_prots_uniq2

        all_prots_uniq = list(set(all_prots))  # there might be repeated prots
        rhogs_prots[host_hog] = all_prots_uniq
        # merging D0562038 and D0559070
        # tr | C3ZG56 | C3ZG56_BRAFL HOG: D0562038, tr | H2Y1V7 | H2Y1V7_CIOIN HOG: D0562038, tr | C3ZG56 | C3ZG56_BRAFL HOG: D0559070, tr | H2Y1V7 | H2Y1V7_CIOIN HOG: D0559070
        if len(all_prots_uniq) > rhogs_host_size:
            counter_merged_prots += len(all_prots_uniq)
        # otherwise, merging didn't help

    print(len(rhogs_prots),counter_merged_prots)

    num_rhog_g1 = 0
    for rhg, list_pr in rhogs_prots.items():
        if len(list_pr) > 1:
            num_rhog_g1 += 1

    logger_hog.debug("There are " + str(num_rhog_g1) + " rhogs (size>1)   by merging "+str(counter_merged_prots)+" proteins in total.")

    return rhogs_prots



def merge_rhogs(hogmaps, rhogs_prots):
    logger_hog.debug("started merging  ")

    candidates_pair = find_rhog_candidate_pairs(hogmaps, rhogs_prots)

    cluster_rhogs_list = cluster_rhogs(candidates_pair)
    num_rhog_g1 = 0
    for rhg, list_pr in rhogs_prots.items():
        if len(list_pr) > 1:
            num_rhog_g1 += 1

    logger_hog.debug("There are " + str(num_rhog_g1) + " rhogs (size>1) before merging.")
    print(len(rhogs_prots))

    file_out_merge =  open("merging_rhogs.txt","w")
    file_out_merge.write("#first column is the host hog, the rest will be merged here.\n")
    for cluster in cluster_rhogs_list:
        cluster_host = min(cluster) # write the host rhog as the first element
        file_out_merge.write(cluster_host + "\t")
        for cluster_i in cluster:
            if cluster_i != cluster_host:
                file_out_merge.write(cluster_i+"\t")
        file_out_merge.write("\n")
    file_out_merge.close()

    counter_merged_prots=0
    newhost= ""
    for cluster in cluster_rhogs_list:
        if len(cluster)>30:
            print(len(cluster))
        #prots = [rhogs_prots[hog] for hog in cluster]
         #  cluster[0] #
        found = False
        while not found and len(cluster)>0:
            host_hog = min(cluster)
            if host_hog in rhogs_prots:
                found = True
                rhogs_host_size = len(rhogs_prots[host_hog])
            else:
                # logger print rhog
                cluster.remove(host_hog)

        all_prots = []
        all_prots2 =[]
        for hog in cluster:
            # todo check, create smallers clusters
            if len(all_prots) < _config.big_rhog_size:
                if hog in rhogs_prots:
                    all_prots += rhogs_prots[hog]
                    del rhogs_prots[hog]
            else:
                if newhost:
                    if len(all_prots2) < _config.big_rhog_size:
                        if hog in rhogs_prots:
                            all_prots2 += rhogs_prots[hog]
                            del rhogs_prots[hog]
                    else:
                        print("cluster is very big "+str(len(cluster)))
                else:
                    newhost = hog

        if newhost:
            all_prots_uniq2 = list(set(all_prots2))  # there might be repeated prots
            rhogs_prots[newhost] = all_prots_uniq2

        all_prots_uniq = list(set(all_prots))  # there might be repeated prots
        rhogs_prots[host_hog] = all_prots_uniq
        # merging D0562038 and D0559070
        # tr | C3ZG56 | C3ZG56_BRAFL HOG: D0562038, tr | H2Y1V7 | H2Y1V7_CIOIN HOG: D0562038, tr | C3ZG56 | C3ZG56_BRAFL HOG: D0559070, tr | H2Y1V7 | H2Y1V7_CIOIN HOG: D0559070
        if len(all_prots_uniq) > rhogs_host_size:
            counter_merged_prots += len(all_prots_uniq)
        # otherwise, merging didn't help

    print(len(rhogs_prots),counter_merged_prots)

    num_rhog_g1 = 0
    for rhg, list_pr in rhogs_prots.items():
        if len(list_pr) > 1:
            num_rhog_g1 += 1

    logger_hog.debug("There are " + str(num_rhog_g1) + " rhogs (size>1)   by merging "+str(counter_merged_prots)+" proteins in total.")

    return rhogs_prots


def collect_unmapped_singleton(rhogs_prots, unmapped,prot_recs_all,unmapped_singleton_fasta= "singleton_unmapped.fa"):
    unmapped_recs = []
    for species_name, prot_names in unmapped.items():
        for prot_name in prot_names:
            prot_rec = prot_recs_all[species_name][prot_name]
            unmapped_recs.append(prot_rec)

    print(len(unmapped_recs))
    singleton_recs = []
    for rhogid, sp_prot_list in rhogs_prots.items():
        if len(sp_prot_list) == 1:
            species_name = sp_prot_list[0][0]
            prot_name = sp_prot_list[0][1]
            prot_rec = prot_recs_all[species_name][prot_name]
            singleton_recs.append(prot_rec)
    print(len(singleton_recs))

    singleton_unmapped_recs = unmapped_recs + singleton_recs

    SeqIO.write(singleton_unmapped_recs, unmapped_singleton_fasta , "fasta")

    return len(singleton_unmapped_recs)


import subprocess


def run_linclust(fasta_to_cluster="singleton_unmapped.fa"):
    num_threads = 10 # todo how to assign more cpu for this step
    command_clust = _config.mmseqs_executable_path +" easy-linclust --threads " + str(
        num_threads) + " " + fasta_to_cluster + " singleton_unmapped tmp_linclust"

    logger_hog.debug("linclust rooting started" + command_clust)
    process = subprocess.Popen(command_clust.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()

    # if verbose:
    #    print("output:\n", output)
    #    print("error:\n", error)

    # if "Error analyzing file" in str(output) or error:
    #    try:

    return "done"


def write_clusters(address_rhogs_folder, min_rhog_size):
    cluster_output_address = "singleton_unmapped_all_seqs.fasta"
    cluster_file = open(cluster_output_address, 'r')
    cluster_dic = {}
    previous_line_start = " "
    # parsing memseqs fasta-like format https://github.com/soedinglab/MMseqs2/wiki#cluster-fasta-like-format
    # >A0A0R4IFW6
    # >tr|A0A0R4IFW6|A0A0R4IFW6_DANRE||DANRE||1000001584 tr|A0A0R
    # MSRKTTSKRHYKPSSEIDDAALARKREYW
    clusters = []
    cluster = []
    line_strip = " "
    # line_number =1
    for line in cluster_file:
        line_strip = line.strip()
        # print(line_number,previous_line_start,line_strip[0],cluster )# "_",previous_line_start,line_strip[0])
        if previous_line_start == " ":
            clusters = []
            cluster = []
        elif line_strip[0] == ">" and previous_line_start == ">":  # new cluster started at previous line
            if len(cluster) >= 5:  # more than one records [ID1, seq1,ID2, seq2, newcluster_id]
                clusters.append(cluster[:-1])  # last item is the id of new cluster
            cluster = [line_strip]
        elif line_strip[0] == ">" and previous_line_start != ">":
            cluster.append(line_strip)
        elif line_strip[0] != ">":
            cluster.append(line_strip)

        previous_line_start = line_strip[0]
        # line_number+=1

    # for last cluster
    if len(cluster) >= 4:  # more than one records  [ID1, seq1,ID2, seq2]
        clusters.append(cluster)

    logger_hog.debug("Number of linclust clusters raw is " + str(len(clusters)))
    # the record id is parsed by mmseqs very ad hoc. so better use the fasta-like file
    # cluster_output_address = "singleton_unmapped_cluster.tsv"
    # cluster_file = open(cluster_output_address, 'r')
    # cluster_dic = {}
    # for line in cluster_file:
    #     line_strip = line.strip()
    #     rep, prot= line_strip.split()
    #     if rep in cluster_dic:
    #         cluster_dic[rep].append(prot) # the frist line includ (rep,rep)
    #     else:
    #         cluster_dic[rep]=[prot]
    # cluster_list = []
    # for rep, prot_list in cluster_dic.items():
    #     if len(prot_list)>1:
    #         cluster_list.append(prot_list)


    cluster_iter= 1000*10
    for cluster in clusters:
        if len(cluster) >= 2 * min_rhog_size:
            species_names= []     # it seems that genes from same species tend to cluster together, discard such clusters
            for pr_idi in  range(int(len(cluster)/2)):
                species_name = cluster[pr_idi*2].split(" ")[0].split("||")[1]
                species_names.append(species_name)
            if len(set(species_names))>1:
                file_idx = open(address_rhogs_folder + "/HOG_clust" + str(cluster_iter) + ".fa", "w")
                for line in cluster:
                    file_idx.write(line + "\n")
                file_idx.close()
                cluster_iter+=1

    return cluster_iter-1


def parse_isoform_file(species_names, folder=None):
    if folder is None:
        folder = "./splice"
    isoform_by_gene_all = {}

    for species_name in species_names:  # from fasta file
        file_splice_name = os.path.join(folder,  species_name + ".splice")
        if os.path.isfile(file_splice_name):
            # from OMArk
            isoform_by_gene = []
            with open(file_splice_name) as handle:
                for line in handle.readlines():
                    line = line.strip('\n')
                    splice = line.split(";")
                    isoform_by_gene.append(splice)
        else:
            isoform_by_gene = []
        isoform_by_gene_all[species_name] = isoform_by_gene
    return isoform_by_gene_all


def find_nonbest_isoform(species_names, isoform_by_gene_all, hogmaps):
    isoform_selected = {}
    isoform_not_selected = {}

    for species_name in species_names:
        isoform_not_selected[species_name] = []
        isoform_selected[species_name] = []
        isoform_list_sp = isoform_by_gene_all[species_name]
        hogmaps_sp = hogmaps[species_name]
        for isoform_list in isoform_list_sp:
            score_isof_best = 0
            protname_best = isoform_list[0]  # to keep the order if all of the isoforms' omamer score are 'na'
            for isoform_name in isoform_list:
                if isoform_name in hogmaps_sp:

                    prot_maps = hogmaps_sp[isoform_name]
                    if len(prot_maps) > 1:  # for multi-hit omamer output with -n
                        scores = [float(i[1]) for i in prot_maps]  # (hogid,score,seqlen,subfamily_medianseqlen)
                        hogids = [i[0] for i in prot_maps]
                        seq_lens = [i[2] for i in prot_maps]
                        subf_med_lens = [i[3] for i in prot_maps]
                        family_score = max(scores)
                        idx_max = scores.index(family_score)
                        seq_len = seq_lens[idx_max]
                        subf_med_len = subf_med_lens[idx_max]
                        # hogid = hogids[max_index]
                    else:
                        # hogid = prot_maps[0][0]
                        family_score = float(prot_maps[0][1])
                        seq_len = prot_maps[0][2]
                        subf_med_len = prot_maps[0][3]
                    score_isof = float(family_score) # for omamer v2 fscore is enough * min(int(seq_len), int(subf_med_len))
                    # print(isoform_name, score_isof,score_isof_best)

                    if score_isof >= score_isof_best:  # when there is a tie, the last one is selected!
                        protname_best = isoform_name
                        score_isof_best = score_isof

            isoform_selected[species_name].append(protname_best)
            isoform_not_selected[species_name] += [i for i in isoform_list if i != protname_best]  # flatten for all

    return isoform_selected, isoform_not_selected


def write_isoform_selected(isoform_by_gene_all, isoform_selected, prot_recs_lists):
    """
    write isoforms

    """
    selected_isoforms_folder = "selected_isoforms/"
    if not os.path.exists(selected_isoforms_folder):
        os.mkdir(selected_isoforms_folder)

    for species, isoform_selected_sp in isoform_selected.items():
        isoform_by_gene = isoform_by_gene_all[species]
        assert len(isoform_by_gene) == len(isoform_selected_sp)

        file_handle = open(selected_isoforms_folder+species + "_selected_isoforms.tsv", "w")
        for gene_idx, isof_list in enumerate(isoform_by_gene):
            isoform_selected_sp_i = isoform_selected_sp[gene_idx]
            file_handle.write(";".join(isof_list) + "\t")
            file_handle.write(isoform_selected_sp_i + "\n")

        all_genes = []
        for list_genes in isoform_by_gene_all[species]:
            all_genes += list_genes
        len(all_genes)

        prot_all = [i.id.split("||")[0] for i in prot_recs_lists[species]]
        len(prot_all), len(all_genes)

        not_in_isoform = set(prot_all) - set(all_genes)

        for isof1 in not_in_isoform:  # for these there is no isoform information, we write them at end of file
            file_handle.write(isof1 + "\t" + isof1 + "\n")

        file_handle.close()

    return 1


def handle_splice(hogmaps,isoform_not_selected):
    hogmaps_selected_isof = {}
    for species, hogmap in hogmaps.items():
        hogmaps_selected_isof[species]={}

        isoform_not_selected_sp = set(isoform_not_selected[species])
        for prot in hogmap:
            if prot not in isoform_not_selected_sp:
                hogmaps_selected_isof[species][prot]=hogmap[prot]

    return hogmaps_selected_isof





def find_outgroup_species(species_tree):
    dic_outgroup_species = {}
    outgroup_species = []

    for node in species_tree.traverse():
        if node.is_leaf() or node.is_root():
            continue

        parent_node = node.up
        outgroup_leaves = []
        if parent_node:
            grandparent_node = parent_node.up
            if grandparent_node:
                # parent_node, grandparent_node,grandparent_node.children
                great_grandparent_node = grandparent_node.up
                if great_grandparent_node:
                    if great_grandparent_node.children:
                        very_distant_cousins = [i for i in great_grandparent_node.children if i != grandparent_node]
                        for i in very_distant_cousins:
                            outgroup_leaves += i.get_leaves()  # a list
                if grandparent_node.children and len(outgroup_leaves) <= 5:
                    distant_cousins = [i for i in grandparent_node.children if i != parent_node]
                    for i in distant_cousins:
                        outgroup_leaves += i.get_leaves()  # a list

            if parent_node.children and len(outgroup_leaves) <= 5:
                cousins = [i for i in parent_node.children if i != node]
                for i in cousins:
                    outgroup_leaves += i.get_leaves()  # a list

        if outgroup_leaves:
            outgroup_species = [i.name for i in outgroup_leaves][:5]

        # print(node.name, outgroup_species)
        dic_outgroup_species[node.name] = outgroup_species

    return dic_outgroup_species


def find_outgroup_prot(dic_outgroup_species, rhog_prot):
    dic_prot_sp = {}
    for sp, prot in rhog_prot:
        if sp in dic_prot_sp:
            dic_prot_sp[sp].append(prot)
        else:
            dic_prot_sp[sp] = [prot]

    dic_outgroup_prot = {}
    for tax, outgroup_species in dic_outgroup_species.items():
        outgroup_prots = []
        for sp in outgroup_species:
            prot_list = dic_prot_sp[sp]
            if len(prot_list) > 1:
                outgroup_prots.append((sp, prot_list[0]))
                outgroup_prots.append((sp, prot_list[1]))
            else:
                outgroup_prots.append((sp, prot_list[0]))

        dic_outgroup_prot[tax] = outgroup_prots

    return dic_outgroup_prot


def write_outgroups(dic_outgroup_prot, rhogid, address_outgorup,prot_recs_all):
    dic_rhog_recs = {}
    for tax, outgroup_prot in dic_outgroup_prot.items():
        rhog_recs = []
        for (species_name, prot_name) in outgroup_prot:
            prot_rec = prot_recs_all[species_name][prot_name]
            rhog_recs.append(prot_rec)
        dic_rhog_recs[tax] = rhog_recs

    for tax, rhog_recs in dic_rhog_recs.items():
        SeqIO.write(rhog_recs, address_outgorup + "/outgroup_" + str(rhogid) + "_" + tax + ".fa", "fasta")

    return 1


def write_outgroups_all(rhogs_prots,prot_recs_all):
    from ete3 import Tree
    sp_folder = "/work/FAC/FBM/DBC/cdessim2/default/smajidi1/qfo/qfo_run_raw/in_folder/"
    address_outgorup = "./outgroup/"
    if not os.path.exists(address_outgorup):
        os.mkdir(address_outgorup)

    print("start working on  outgroup generation for rhogs  "+str(len(rhogs_prots)))
    for rhog, rhog_prot in rhogs_prots.items():
        if len(rhog_prot) > 2:
            # print(rhog)
            species_tree = Tree(sp_folder + "species_tree.nwk", format=1)
            list_species = list(set([i[0] for i in rhog_prot]))
            species_tree.prune(list_species)
            # species_tree.write(format=1)
            # if len(species_tree)>2:
            #    counter+=1
            dic_outgroup_species = find_outgroup_species(species_tree)
            dic_outgroup_species

            dic_outgroup_prot = find_outgroup_prot(dic_outgroup_species, rhog_prot)

            # print(len(dic_outgroup_prot))

            write_outgroups(dic_outgroup_prot, rhog, address_outgorup,prot_recs_all)
    print("done writing")
    return 1



def cluster_rhogs_nx(cluster_rhogs_list,candidates_pair):
    count=0
    candidates_pair = set(candidates_pair)
    new_cluster = []
    for c in cluster_rhogs_list:
        count+=1

        G = nx.Graph()
        for i in range(len(c)):
            pairA = c[i]

            if not G.has_node(pairA):
                G.add_node(pairA)
            for j in range(i+1, len(c)):

                pairB = c[j]
                if not G.has_node(pairB):
                    G.add_node(pairB)
                if (pairA,pairB) in candidates_pair or (pairB,pairA) in candidates_pair:
                    G.add_edge(pairA,pairB)
        density = nx.density(G)

        newG = HCS(G)
        clusters = [ list(x) for x in nx.connected_components(newG) if len(x)>1]
        new_cluster += clusters
    return new_cluster

def highly_connected(G, E):
    """Checks if the graph G is highly connected

    Highly connected means, that splitting the graph G into subgraphs needs more than 0.5*|V| edge deletions
    This definition can be found in Section 2 of the publication.

    :param G: Graph G
    :param E: Edges needed for splitting G
    :return: True if G is highly connected, otherwise False
    """

    return len(E) >= len(G.nodes) / 2


def remove_edges(G, E):
    """Removes all edges E from G

    Iterates over all edges in E and removes them from G
    :param G: Graph to remove edges from
    :param E: One or multiple Edges
    :return: Graph with edges removed
    """

    for edge in E:
        G.remove_edge(*edge)
    return G


def HCS(G):
    """Basic HCS Algorithm
https://github.com/53RT/Highly-Connected-Subgraphs-Clustering-HCS/blob/master/hcs.py
    cluster labels, removed edges are stored in global variables

    :param G: Input graph
    :return: Either the input Graph if it is highly connected, otherwise a Graph composed of
    Subgraphs that build clusters
    """

    E = nx.algorithms.connectivity.cuts.minimum_edge_cut(G)

    if not highly_connected(G, E):
        G = remove_edges(G, E)
        sub_graphs = [G.subgraph(c).copy() for c in nx.connected_components(G)]

        if len(sub_graphs) == 2:
            H = HCS(sub_graphs[0])
            _H = HCS(sub_graphs[1])

            G = nx.compose(H, _H)

    return G




import numpy as np
from Bio.Seq import Seq  # , UnknownSeq
from Bio.Align import MultipleSeqAlignment
from Bio.SeqRecord import SeqRecord
from ete3 import Tree


from . import _wrappers
from . import _utils_subhog
from . import _config
from ._config import logger_hog


"""
This code is for detecting fragments in MSA due to poor gene annotation and used a merged version in orthology relationship "_|_" in the gen trees.
reported in orthoxml either as DubiousMergedfragment or Dubiousfragment

We also remove proteins with low SO (species_overlap) score. 
30 Apr 2022
"""

def read_msa(input_msa):
    ids = []
    seqs = []
    species = []
    coords = []
    for rec in input_msa:
        ids.append(rec.id)
        seqs.append(np.frombuffer(str(rec.seq).upper().encode('ascii'), dtype='S1'))
        species.append(rec.id.split('||')[1][:-1])

        # compute coordinates in matrix
        # todo  if all the element is --, this will be empty and causing error
        ii = np.argwhere(seqs[-1] != b'-')
        if len(ii):
            coords.append((np.min(ii), np.max(ii)))
        else:
            logger_hog.warning("issue 1321230 all of the seq is gap.  "+str(rec))
            coords.append((0, 0))

    ids = np.array(ids)
    seqs = np.vstack(seqs)
    species = np.array(species)
    coords = np.array(coords)
    input_msa_np = (ids, seqs, species, coords)
    return input_msa_np



def compute_identity(s1, s2):
    def compute(s1, s2):
        # ignore gap columns in first sequence
        f = (s1 != b'-')
        n = f.sum()
        return ((s1[f] == s2[f]).sum() / n) if n > 0 else 1

    return compute(s1, s2), compute(s2, s1)


def split_candidates(input_msa_np, margin=0):


    (ids_all, seqs_all, species_all, coords_all) = input_msa_np
    '''
    Count candidates where it is unambiguous which fragments to merge.
    by Alex Warwick Vesztrocy
    '''
    assert 0 <= margin <= 1, "Margin must be in [0,1]"
    candidates = []

    # get species counts
    c = np.unique(species_all, return_counts=True)
    total_seqs = c[1].sum()

    # filter to those with more than 1 seq
    f = (c[1] > 1)
    for sp in c[0][f]:
        ii = np.argwhere(species_all == sp).flatten()
        # print('species', sp, margin, len(ii), total_seqs)
        x = coords_all[ii]

        order = np.argsort(x[:, 0])
        coords = x[order]
        ids = ids_all[ii[order]]
        seqs = seqs_all[ii[order]]

        #  find whether to merge for unambiguous cases
        merge = True
        ident = []
        for i in range(len(coords) - 1):
            (s1, e1) = coords[i]
            (s2, e2) = coords[i + 1]
            i_margin = min(margin * (e1 - s1),
                           margin * (e2 - s2))
            # s1-------e1
            #       s2------e2
            #  (e1-s2) < 0 no overlap; (e1-s2) > 0 overlap.
            overlap = (e1 - s2)
            # print(overlap)
            # todo  we could improve finding overlap between fragments
            if overlap < 0:
                ident.append((1, 1))
            elif overlap <= i_margin:
                # minimal overlap, compute identities
                ident.append(
                    compute_identity(
                        seqs[i][s2:e1],
                        seqs[i + 1][s2:e1],
                    )
                )
            else:
                # overlapping too much
                merge = False
                break
        if merge:
            candidates.append((tuple(ids), ident))
    return candidates


def find_prot_dubious_msa(input_msa):

    input_msa_np = read_msa(input_msa)
    candidates = split_candidates(input_msa_np, _config.overlap_fragments)

    # candidates.append((tuple(ids), ident))
    #prot_dubious_msa_list = [(str(i[0][0]), str(i[0][1])) for i in candidates]  # [i[0] for i in candidates]
    # todo are there only two always? few fragments
    prot_dubious_msa_list = []
    seq_dubious_msa_list = []

    if candidates :
        for candidate in candidates:
            prot_dubious_msa_list.append(list(candidate[0]))

        for prots_dubious_msa in prot_dubious_msa_list:
            seqs_dubious_msa = [i for i in input_msa if i.name in prots_dubious_msa]
            seq_dubious_msa_list.append(seqs_dubious_msa)

    return prot_dubious_msa_list,  seq_dubious_msa_list  # list of pairs


def insert_dubious_prots_hog_hierarchy_toleaves(hog_host, fragment_host, fragments_list_nothost):
    for subhog in hog_host._subhogs:

        insert_dubious_prots_hog_hierarchy_toleaves(subhog, fragment_host, fragments_list_nothost)

    result_insersion = hog_host.insert_dubious_prots(fragment_host, fragments_list_nothost)

    return 1



def remove_prot_hog_hierarchy_toleaves(hog_ii, prot_to_remove):
    # import copy; a= copy.deepcopy(hog_ii)
    for subhog in hog_ii._subhogs:
        if prot_to_remove in subhog._members:
            result_removing0 = remove_prot_hog_hierarchy_toleaves(subhog, prot_to_remove)

    result_removing = hog_ii.remove_prot_from_hog(prot_to_remove)
    subhogs_old = hog_ii._subhogs
    hog_ii._subhogs = [i for i in subhogs_old if len(i._members)]  # for getting rid of  empty subhogs #  hogID=HOG_0026884_sub12868,len=0, tax_least=Ardeidae

    return result_removing



def find_prot_dubious_sd_remove(gene_tree, all_species_dubious_sd_dic):
    # todo this function need to double check with cases of with and without dubious

    #prot_dubious_sd_allspecies = []
    prot_dubious_sd_remove_list = []
    # todo not sure postorder or preorder
    for node in gene_tree.traverse(strategy="postorder"):
        # print("** now working on node ",node.name) # node_children
        if not node.is_leaf() and 'D' in node.name:
            node_name = node.name #d, intersection, union = node_name.split("_")  # if int(intersection) / int(union) < _config.threshold_dubious_sd:
            if node_name in all_species_dubious_sd_dic: # a duplication node with low score,
                node_children = node.children
                all_species_dubious_sd = all_species_dubious_sd_dic[node_name]
                # prot_dubious_sd_ = []
                for species_dubious_sd in all_species_dubious_sd:
                    child_size = []  # gene tree is binary for fasttree
                    prot_dubious_list = []
                    for node_child in node_children:
                        list_leaves = [i.name for i in node_child.get_leaves()]
                        child_size.append(len(list_leaves))
                        for prot_name in list_leaves:
                            if prot_name.split("||")[1] == species_dubious_sd:
                                prot_dubious_list.append(prot_name)
                    #try:
                    subhogs_list = [i.split("|_|")[1][:-1] for i in prot_dubious_list]  # subhog id at child level
                    #  ["'tr|Q4V8S5|Q4V8S5_DANRE||DANRE||1000020519|_|sub12962'"]
                    # the last char is ', watch out!

                    # todo check ! is it safe or not!
                    # if len(set(subhogs_list)) > 1:
                    #     # we are removing all sequences of this species on the the side of internal node (gene tree), with least leaves
                    #     child_size_min_indx = child_size.index(min(child_size))
                    #     prot_dubious_sd_remove_list.append(prot_dubious_list[child_size_min_indx])
                    #
                    # else:
                    #     logger_hog.debug( "This species (protein from the same subhog) is safe to keep "+ str(node_name)+" "+str(species_dubious_sd))
                    #     #all of them are from the same subhog, so it doesn't matter, a duplication event doesn't affect when all are from the same subhog at children level

                    child_size_min_indx = child_size.index(min(child_size))
                    prot_dubious_sd_remove_list.append(prot_dubious_list[child_size_min_indx]) # ["'sp|Q9PRL8|ACBP_CHICK||CHICK||1020017457|_|sub10101'"]

                    # except:
                    #    logger_hog.warning("issue 2495869: prot_dubious_list doesnt include the hog id . so we'll  keep it" + str(prot_dubious_list)+ " " +str(gene_tree.write(format=1, format_root_node=True) ) )


    return prot_dubious_sd_remove_list




def handle_fragment_sd(node_species_tree, gene_tree, genetree_msa_file_addr, all_species_dubious_sd_dic, hogs_children_level_list):
    #  prot_dubious_sd_list, node_species_tree, genetree_msa_file_addr, hogs_children_level_list

    all_species_dubious_sd_dic_updated = all_species_dubious_sd_dic
    itr_so = 0
    while all_species_dubious_sd_dic_updated and itr_so< 10: # todo future parameter 10
        logger_hog.debug("These are found with low SO score all_species_dubious_sd_dic " + str(all_species_dubious_sd_dic_updated)+" which are now being handled itr"+str(itr_so)+" .")
        prot_dubious_sd_remove_list = find_prot_dubious_sd_remove(gene_tree, all_species_dubious_sd_dic_updated)
        if prot_dubious_sd_remove_list:
            rest_leaves = set([i.name for i in gene_tree.get_leaves()]) - set(prot_dubious_sd_remove_list)
            gene_tree.prune(rest_leaves, preserve_branch_length=True)
            if _config.gene_trees_write_all or _config.rooting_method=="mad":
                gene_tree.write(format=1, outfile=genetree_msa_file_addr + "_dubious_sd"+str(itr_so)+".nwk" )

            (gene_tree, all_species_dubious_sd_dic_updated) = _utils_subhog.genetree_sd(node_species_tree, gene_tree, genetree_msa_file_addr + "_dubious_sd"+str(itr_so)+".nwk")

            hogs_children_level_list_raw = hogs_children_level_list
            for prot_dubious_sd_remove in prot_dubious_sd_remove_list:
                logger_hog.debug("** we are removing the sequence "+str(prot_dubious_sd_remove)+"due to low species overlap score")

            for prot_dubious_sd_remove in prot_dubious_sd_remove_list:
                for hog in hogs_children_level_list_raw:
                    if prot_dubious_sd_remove in hog._members:
                        result_removing = remove_prot_hog_hierarchy_toleaves(hog, prot_dubious_sd_remove)
                        if result_removing == 0:  # the hog is empty
                            hogs_children_level_list.remove(hog)

        else:
            all_species_dubious_sd_dic_updated = []

        itr_so += 1

    return (gene_tree, hogs_children_level_list)


def merge_prots_name_hierarchy_toleaves(hog_host, fragment_name_host, merged_fragment_name):

    # todo do I really need this ? as I don't check the members of those subhogs close to the leaves anymore.. to keep it consistent
    for subhog in hog_host._subhogs:
        if fragment_name_host in subhog._members:
            merge_prots_name_hierarchy_toleaves(subhog, fragment_name_host, merged_fragment_name)

    result_merging = hog_host.merge_prots_name_hog(fragment_name_host, merged_fragment_name)

    return hog_host


def merge_fragments_hogclass(fragments_set, seq_dubious_msa, hogs_children_level_list_raw, merged_msa):
    # a bit of redundancy fragments_set is the names which also are stored in  seq_dubious_msa

    fragments_list = list(fragments_set)
    fragment_name_host = fragments_list[0]
    for hog in hogs_children_level_list_raw:
        if fragment_name_host in hog._members:
            hog_host = hog
    merged_fragment_name = "_|_".join(fragments_list)  # fragment_name_host + "_|_" + fragments_list
    fragments_list_remove = fragments_list[1:]
    hogs_to_remove_list =[]
    for fragment_idx in range(1, len(fragments_list)):  # the 0 element is the host

        fragment_name_remove = fragments_list[fragment_idx]
        # fragment_seq_remove = seq_dubious_msa[fragment_idx]
        # remove
        for hog in hogs_children_level_list_raw:
            if fragment_name_remove in hog._members:
                result_removing = remove_prot_hog_hierarchy_toleaves(hog, fragment_name_remove)
                if result_removing == 0:  # the hog is empty
                    hogs_to_remove_list.append(hog)

    hogs_children_level_list = [i for i in hogs_children_level_list_raw if i not in hogs_to_remove_list]

    merged_sequence = ""
    for column_idx in range(len(seq_dubious_msa[0])):
        # amino-acid
        aa_col_list = [str(rec.seq)[column_idx] for rec in seq_dubious_msa]
        aa_col_set = set(aa_col_list)
        if aa_col_set == {'-'}:
            aa_consensus='-'
        elif len(aa_col_set-{'-'}) == 1:
            aa_consensus = next(iter(aa_col_set-{'-'}))
        elif len(aa_col_set-{'-'}) > 1:
            aa_consensus = 'X'
        else:
            logger_hog.WARNING("issue 123124124"+str(aa_col_set))

        merged_sequence += aa_consensus
    # seq0 = str(seq_dubious_msa[0].seq)
    # seq1 = str(seq_dubious_msa[1].seq)
    # assert len(seq0) == len(seq1)
    # merged_sequence = ""
    # for aa_idx in range(len(seq1)):
    #     if seq0[aa_idx] == '-':
    #         merged_sequence += seq1[aa_idx]
    #     elif seq1[aa_idx] == '-':
    #         merged_sequence += seq0[aa_idx]
    #     elif seq0[aa_idx] == seq1[aa_idx]:   #  and seq0 != '-' and seq1 != '-'
    #         merged_sequence += seq0[aa_idx]
    #     else:
    #         merged_sequence += "X"
    # assert len(merged_sequence) == len(seq0)
    # merged_fragment_name = "_|_" .join(fragments_list) # fragment_name_host + "_|_" + fragments_list
    if len(merged_fragment_name) > 220:
        logger_hog.warning("The length of sequence id which now being merged is getting very long > 220, we should make sure that it won't cause any issues with fasttree nd biopython and Mafft"+str(merged_fragment_name))
    merged_msa_new_list = []
    if merged_msa and merged_msa[0]:
        assert len(merged_msa[0]) == len(merged_sequence), str(fragment_name_host)
    else:
        logger_hog.warning("issue 15723 merged_msa is empty ?")
    for seq_rec in merged_msa:
        if seq_rec.id == fragment_name_host:
            seq_rec_edited = SeqRecord(Seq(merged_sequence), id=merged_fragment_name, name=merged_fragment_name)
            merged_msa_new_list.append(seq_rec_edited)
        elif seq_rec.id in fragments_list_remove:
            pass
        else:
            merged_msa_new_list.append(seq_rec)
    merged_msa_new = MultipleSeqAlignment(merged_msa_new_list)

    # fragments_list_remove
    result_merging1 = merge_prots_name_hierarchy_toleaves(hog_host, fragment_name_host, merged_fragment_name)

    #hog_host.merge_prots_msa(fragment_name_host, fragment_name_remove, merged_sequence, merged_msa_new)
    hog_host.merge_prots_msa(merged_fragment_name, merged_msa_new)
    logger_hog.debug("these proteins fragments are merged  into one "+str(merged_fragment_name)+" but reported seperately in orthoxml")


    return fragments_list_remove, hogs_children_level_list, merged_msa_new



def handle_fragment_msa(prot_dubious_msa_list, seq_dubious_msa_list, gene_tree, node_species_tree, genetree_msa_file_addr, hogs_children_level_list, merged_msa):
    merged_msa_new = merged_msa
    if not prot_dubious_msa_list:  # empty list
        return gene_tree, hogs_children_level_list, merged_msa_new

    logger_hog.debug("** these are found prot_dubious_msa_list " + str(prot_dubious_msa_list))
    fragments_set_list, seq_dubious_msa_list_checked = check_prot_dubious_msa(prot_dubious_msa_list, gene_tree, seq_dubious_msa_list)
    fragments_remove_list = [] # this list include only one of the fragments for each set, the other one is merged version afterwards
    if fragments_set_list and len(fragments_set_list[0]) > 1:
        # logger_hog.debug("** these are found fragments_set_list " + str(fragments_set_list))
        # remove fragments from gene tree
        if _config.fragment_detection and _config.fragment_detection_msa_merge:
            merged_msa_new = merged_msa
            for fragments_set_idx, fragments_set in enumerate(fragments_set_list):
                fragments_set = fragments_set_list[fragments_set_idx]
                seq_dubious_msa = seq_dubious_msa_list_checked[fragments_set_idx]
                fragments_list_remove, hogs_children_level_list, merged_msa_new = merge_fragments_hogclass(fragments_set, seq_dubious_msa, hogs_children_level_list, merged_msa_new)
                # merged_msa_new  contains the merged_seq
                # note that merged_msa is the merging of different subhog childrend and is not related to merge fragments  of two seqeuncs with bad annotation

            if merged_msa_new:
                (msa_filt_row_col_new, msa_filt_col, hogs_children_level_list) = _utils_subhog.filter_msa(merged_msa_new, genetree_msa_file_addr+"_merged", hogs_children_level_list)
            if len(msa_filt_row_col_new) > 1 and len(msa_filt_row_col_new[0]) > 3:
                gene_tree_raw = _wrappers.infer_gene_tree(msa_filt_row_col_new, genetree_msa_file_addr+"_merged_")
                gene_tree = Tree(gene_tree_raw , format=0, quoted_node_names=True) #+ ";"
                if _config.add_outgroup:
                    species_this_node = [i.name for i in node_species_tree.get_leaves()]
                    gene_names = [i.name for i in gene_tree.get_leaves()]
                    gene_names_good = [i for i in gene_names if i.split("||")[1] in species_this_node]
                    gene_tree.prune(gene_names_good, preserve_branch_length=True)

            else:
                logger_hog.warning("** issue 861956")
                gene_tree = ""
                # fragments_remove_list += fragments_list_remove # for now fragments_list_remove include 1 prots
        elif _config.fragment_detection and (not _config.fragment_detection_msa_merge):
            fragments_remove_set = set.union(*fragments_set_list)
            rest_leaves = set([i.name for i in gene_tree.get_leaves()]) - fragments_remove_set
            if len(rest_leaves) < 2:
                # todo
                logger_hog.warning("** issue 86194")
                hogs_children_level_list = []
                gene_tree = ""
                return gene_tree, hogs_children_level_list, merged_msa_new
            else:
                gene_tree.prune(rest_leaves, preserve_branch_length=True)

        try:
            if  len(gene_tree) and (_config.gene_trees_write_all or _config.rooting_method == "mad"): # len(gene_tree) > 1 and
                gene_tree.write(outfile=genetree_msa_file_addr+"_dubiousMSA.nwk",format=1)
        except:
            logger_hog.warning("couldn't write the file _dubiousMSA.nwk")


        if len(gene_tree) > 1:
            (gene_tree, all_species_dubious_sd_dic2) = _utils_subhog.genetree_sd(node_species_tree, gene_tree, genetree_msa_file_addr+"_dubiousMSA.nwk")


        # todo check the following is needed
        # if all_species_dubious_sd_dic2:
        #     # logger_hog.debug("these are  found after removing with msa , all_species_dubious_sd_dic2 "+str(all_species_dubious_sd_dic2))
        #     (gene_tree, hogs_children_level_list) = handle_fragment_sd(node_species_tree, gene_tree, genetree_msa_file_addr, all_species_dubious_sd_dic2, hogs_children_level_list)
        # the handle_fragment_sd might have the issue as the gene tree might not contain the subHOG ids

        # todo the following was part of this if condition. are they needed
        # for fragments_set in fragments_set_list:
        #     fragments_list = list(fragments_set)
        #     fragment_host = fragments_list[0]  # host of the new small hog consisting few fragments
        #     for hog in hogs_children_level_list:
        #         if fragment_host in hog._members:
        #             hog_host = hog
        #             break  # once we found, we don't need to continue searching in hog
        #     fragments_list_nothost = fragments_list[1:]
        #     for fragment in fragments_list_nothost:
        #         hogs_children_level_list_raw = hogs_children_level_list
        #         for hog in hogs_children_level_list_raw:
        #             if fragment in hog._members:
        #                 result_removing = remove_prot_hog_hierarchy_toleaves(hog, fragment)
        #                 if result_removing == 0:  # the hog is empty
        #                     hogs_children_level_list.remove(hog)
        #                     # print(hogs_children_level_list)
        #     insert_dubious_prots_hog_hierarchy_toleaves(hog_host, fragment_host, fragments_list_nothost)

    return gene_tree, hogs_children_level_list, merged_msa_new

def check_prot_dubious_msa(prot_dubious_msa_list, gene_tree, seq_dubious_msa_list):

    farthest, max_dist_numNodes = gene_tree.get_farthest_node(topology_only=True)  # furthest from the node
    farthest, max_dist_length = gene_tree.get_farthest_node()  # furthest from the node
    print("max_dist_numNodes, max_dist_length ", max_dist_numNodes, max_dist_length)
    fragments_set_list = []
    # some of the genes may not be in the merged_msa because of trimming rows of msa, we are not using merged_msa
    gene_tree_leaves_name = set([i.name for i in gene_tree.get_leaves()])


    for prot_dubious_msa_idx , prot_dubious_msa_set in enumerate(prot_dubious_msa_list):
        # print(prot_dubious_msa_set)
        fragments = []
        prot_dubious_msa_set_edited = [i for i in prot_dubious_msa_set if i in gene_tree_leaves_name]
        if len(prot_dubious_msa_set_edited)>1:
            prot_host = prot_dubious_msa_set_edited[0]
            for prot in prot_dubious_msa_set_edited[1:]:
                # todo following could be imporved, during filtering row/col msa, a fragments could be removed and not in gene tree anymore,
                # there might be few fragments, checking the distance of the first one with the rest
                # todo this could be improved check all vs all
                assert len(gene_tree.get_leaves_by_name(prot_host)) == 1, "the prot name is not in gene tree or there are more than one" + str(prot_host)
                assert len(gene_tree.get_leaves_by_name(prot)) == 1, "the prot name is not in gene tree or there are more than one" + str(prot)

                # dist_numNodes = gene_tree.get_distance(prot_host, prot, topology_only=True)

                node_prot = gene_tree.get_leaves_by_name(prot)[0]
                node_prot_host = gene_tree.get_leaves_by_name(prot_host)[0]
                dist_length = gene_tree.get_distance(node_prot_host, node_prot)

                dist_length_corrected = dist_length - abs(node_prot.dist- node_prot_host.dist)

                print("check_prot_dubious_msa dist_numNodes, dist_length ", dist_length)
                if dist_length_corrected < max(0.005, max_dist_length / 5):   # or (dist_length_corrected - 2*max_dist_length)< 0.001
                    # todo important hazard to test
                    # dist_numNodes < max(max_dist_numNodes * 1 / 5, 3) or
                    fragments.append(prot)

            if fragments:
                fragments.append(prot_host)
                fragments_set_list.append(set(fragments))

    seq_dubious_msa_list_checked = []
    for fragments_set in fragments_set_list:
        seq_dubious_msa = []
        for seq_original in seq_dubious_msa_list:
            seq_dubious_msa = [i for i in seq_original if i.id in fragments_set]
            if seq_dubious_msa:
                seq_dubious_msa_list_checked.append(seq_dubious_msa)

    logger_hog.debug("These are fragments found " + str(fragments_set_list))

    return fragments_set_list, seq_dubious_msa_list_checked




import abc
import re
from ._config import logger_hog as logger


class FastaHeaderTransformer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def transform(self, header):
        return header


class NoOpFastaHeaderTransformer(FastaHeaderTransformer):
    def transform(self, header):
        return header


class ExtractUniProtAccessionFastaHeaderTransformer(FastaHeaderTransformer):
    def __init__(self):
        self._up_re = re.compile(r"[sptr]{2}\|(?P<acc>[OPQ][0-9][A-Z0-9]{3}[0-9]|[A-NR-Z][0-9]([A-Z][A-Z0-9]{2}[0-9]){1,2})\|.*")

    def transform(self, header):
        m = self._up_re.match(header)
        if m:
            return m.group('acc')
        logger.warning("cannot extract uniprot accession from header: %s", header)
        return header


def header_transformer(name):
    if name.lower() == "noop":
        return NoOpFastaHeaderTransformer()
    elif name.lower() == 'uniprot':
        return ExtractUniProtAccessionFastaHeaderTransformer()
